"""aequitas

Revision ID: b4d7569d31cb
Revises: 609c7cc51794
Create Date: 2019-05-07 11:56:03.814097

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b4d7569d31cb'
down_revision = '609c7cc51794'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('aequitas',
    sa.Column('model_id', sa.Integer(), nullable=False),
    sa.Column('subset_hash', sa.String(), nullable=False),
    sa.Column('tie_breaker', sa.String(), nullable=False),
    sa.Column('evaluation_start_time', sa.DateTime(), nullable=False),
    sa.Column('evaluation_end_time', sa.DateTime(), nullable=False),
    sa.Column('matrix_uuid', sa.Text(), nullable=True),
    sa.Column('parameter', sa.String(), nullable=False),
    sa.Column('attribute_name', sa.String(), nullable=False),
    sa.Column('attribute_value', sa.String(), nullable=False),
    sa.Column('total_entities', sa.Integer(), nullable=True),
    sa.Column('group_label_pos', sa.Integer(), nullable=True),
    sa.Column('group_label_neg', sa.Integer(), nullable=True),
    sa.Column('group_size', sa.Integer(), nullable=True),
    sa.Column('group_size_pct', sa.Numeric(), nullable=True),
    sa.Column('prev', sa.Numeric(), nullable=True),
    sa.Column('pp', sa.Integer(), nullable=True),
    sa.Column('pn', sa.Integer(), nullable=True),
    sa.Column('fp', sa.Integer(), nullable=True),
    sa.Column('fn', sa.Integer(), nullable=True),
    sa.Column('tn', sa.Integer(), nullable=True),
    sa.Column('tp', sa.Integer(), nullable=True),
    sa.Column('ppr', sa.Numeric(), nullable=True),
    sa.Column('pprev', sa.Numeric(), nullable=True),
    sa.Column('tpr', sa.Numeric(), nullable=True),
    sa.Column('tnr', sa.Numeric(), nullable=True),
    sa.Column('for', sa.Numeric(), nullable=True),
    sa.Column('fdr', sa.Numeric(), nullable=True),
    sa.Column('fpr', sa.Numeric(), nullable=True),
    sa.Column('fnr', sa.Numeric(), nullable=True),
    sa.Column('npv', sa.Numeric(), nullable=True),
    sa.Column('precision', sa.Numeric(), nullable=True),
    sa.Column('ppr_disparity', sa.Numeric(), nullable=True),
    sa.Column('ppr_ref_group_value', sa.String(), nullable=True),
    sa.Column('pprev_disparity', sa.Numeric(), nullable=True),
    sa.Column('pprev_ref_group_value', sa.String(), nullable=True),
    sa.Column('precision_disparity', sa.Numeric(), nullable=True),
    sa.Column('precision_ref_group_value', sa.String(), nullable=True),
    sa.Column('fdr_disparity', sa.Numeric(), nullable=True),
    sa.Column('fdr_ref_group_value', sa.String(), nullable=True),
    sa.Column('for_disparity', sa.Numeric(), nullable=True),
    sa.Column('for_ref_group_value', sa.String(), nullable=True),
    sa.Column('fpr_disparity', sa.Numeric(), nullable=True),
    sa.Column('fpr_ref_group_value', sa.String(), nullable=True),
    sa.Column('fnr_disparity', sa.Numeric(), nullable=True),
    sa.Column('fnr_ref_group_value', sa.String(), nullable=True),
    sa.Column('tpr_disparity', sa.Numeric(), nullable=True),
    sa.Column('tpr_ref_group_value', sa.String(), nullable=True),
    sa.Column('tnr_disparity', sa.Numeric(), nullable=True),
    sa.Column('tnr_ref_group_value', sa.String(), nullable=True),
    sa.Column('npv_disparity', sa.Numeric(), nullable=True),
    sa.Column('npv_ref_group_value', sa.String(), nullable=True),
    sa.Column('Statistical_Parity', sa.Boolean(), nullable=True),
    sa.Column('Impact_Parity', sa.Boolean(), nullable=True),
    sa.Column('FDR_Parity', sa.Boolean(), nullable=True),
    sa.Column('FPR_Parity', sa.Boolean(), nullable=True),
    sa.Column('FOR_Parity', sa.Boolean(), nullable=True),
    sa.Column('FNR_Parity', sa.Boolean(), nullable=True),
    sa.Column('TypeI_Parity', sa.Boolean(), nullable=True),
    sa.Column('TypeII_Parity', sa.Boolean(), nullable=True),
    sa.Column('Equalized_Odds', sa.Boolean(), nullable=True),
    sa.Column('Unsupervised_Fairness', sa.Boolean(), nullable=True),
    sa.Column('Supervised_Fairness', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['matrix_uuid'], ['model_metadata.matrices.matrix_uuid'], ),
    sa.ForeignKeyConstraint(['model_id'], ['model_metadata.models.model_id'], ),
    sa.PrimaryKeyConstraint('model_id', 'subset_hash', 'tie_breaker', 'evaluation_start_time', 'evaluation_end_time', 'parameter', 'attribute_name', 'attribute_value'),
    schema='test_results'
    )
    op.create_index(op.f('ix_test_results_aequitas_attribute_name'), 'aequitas', ['attribute_name'], unique=False, schema='test_results')
    op.create_index(op.f('ix_test_results_aequitas_attribute_value'), 'aequitas', ['attribute_value'], unique=False, schema='test_results')
    op.create_index(op.f('ix_test_results_aequitas_evaluation_end_time'), 'aequitas', ['evaluation_end_time'], unique=False, schema='test_results')
    op.create_index(op.f('ix_test_results_aequitas_evaluation_start_time'), 'aequitas', ['evaluation_start_time'], unique=False, schema='test_results')
    op.create_index(op.f('ix_test_results_aequitas_model_id'), 'aequitas', ['model_id'], unique=False, schema='test_results')
    op.create_index(op.f('ix_test_results_aequitas_parameter'), 'aequitas', ['parameter'], unique=False, schema='test_results')
    op.create_table('aequitas',
    sa.Column('model_id', sa.Integer(), nullable=False),
    sa.Column('subset_hash', sa.String(), nullable=False),
    sa.Column('tie_breaker', sa.String(), nullable=False),
    sa.Column('evaluation_start_time', sa.DateTime(), nullable=False),
    sa.Column('evaluation_end_time', sa.DateTime(), nullable=False),
    sa.Column('matrix_uuid', sa.Text(), nullable=True),
    sa.Column('parameter', sa.String(), nullable=False),
    sa.Column('attribute_name', sa.String(), nullable=False),
    sa.Column('attribute_value', sa.String(), nullable=False),
    sa.Column('total_entities', sa.Integer(), nullable=True),
    sa.Column('group_label_pos', sa.Integer(), nullable=True),
    sa.Column('group_label_neg', sa.Integer(), nullable=True),
    sa.Column('group_size', sa.Integer(), nullable=True),
    sa.Column('group_size_pct', sa.Numeric(), nullable=True),
    sa.Column('prev', sa.Numeric(), nullable=True),
    sa.Column('pp', sa.Integer(), nullable=True),
    sa.Column('pn', sa.Integer(), nullable=True),
    sa.Column('fp', sa.Integer(), nullable=True),
    sa.Column('fn', sa.Integer(), nullable=True),
    sa.Column('tn', sa.Integer(), nullable=True),
    sa.Column('tp', sa.Integer(), nullable=True),
    sa.Column('ppr', sa.Numeric(), nullable=True),
    sa.Column('pprev', sa.Numeric(), nullable=True),
    sa.Column('tpr', sa.Numeric(), nullable=True),
    sa.Column('tnr', sa.Numeric(), nullable=True),
    sa.Column('for', sa.Numeric(), nullable=True),
    sa.Column('fdr', sa.Numeric(), nullable=True),
    sa.Column('fpr', sa.Numeric(), nullable=True),
    sa.Column('fnr', sa.Numeric(), nullable=True),
    sa.Column('npv', sa.Numeric(), nullable=True),
    sa.Column('precision', sa.Numeric(), nullable=True),
    sa.Column('ppr_disparity', sa.Numeric(), nullable=True),
    sa.Column('ppr_ref_group_value', sa.String(), nullable=True),
    sa.Column('pprev_disparity', sa.Numeric(), nullable=True),
    sa.Column('pprev_ref_group_value', sa.String(), nullable=True),
    sa.Column('precision_disparity', sa.Numeric(), nullable=True),
    sa.Column('precision_ref_group_value', sa.String(), nullable=True),
    sa.Column('fdr_disparity', sa.Numeric(), nullable=True),
    sa.Column('fdr_ref_group_value', sa.String(), nullable=True),
    sa.Column('for_disparity', sa.Numeric(), nullable=True),
    sa.Column('for_ref_group_value', sa.String(), nullable=True),
    sa.Column('fpr_disparity', sa.Numeric(), nullable=True),
    sa.Column('fpr_ref_group_value', sa.String(), nullable=True),
    sa.Column('fnr_disparity', sa.Numeric(), nullable=True),
    sa.Column('fnr_ref_group_value', sa.String(), nullable=True),
    sa.Column('tpr_disparity', sa.Numeric(), nullable=True),
    sa.Column('tpr_ref_group_value', sa.String(), nullable=True),
    sa.Column('tnr_disparity', sa.Numeric(), nullable=True),
    sa.Column('tnr_ref_group_value', sa.String(), nullable=True),
    sa.Column('npv_disparity', sa.Numeric(), nullable=True),
    sa.Column('npv_ref_group_value', sa.String(), nullable=True),
    sa.Column('Statistical_Parity', sa.Boolean(), nullable=True),
    sa.Column('Impact_Parity', sa.Boolean(), nullable=True),
    sa.Column('FDR_Parity', sa.Boolean(), nullable=True),
    sa.Column('FPR_Parity', sa.Boolean(), nullable=True),
    sa.Column('FOR_Parity', sa.Boolean(), nullable=True),
    sa.Column('FNR_Parity', sa.Boolean(), nullable=True),
    sa.Column('TypeI_Parity', sa.Boolean(), nullable=True),
    sa.Column('TypeII_Parity', sa.Boolean(), nullable=True),
    sa.Column('Equalized_Odds', sa.Boolean(), nullable=True),
    sa.Column('Unsupervised_Fairness', sa.Boolean(), nullable=True),
    sa.Column('Supervised_Fairness', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['matrix_uuid'], ['model_metadata.matrices.matrix_uuid'], ),
    sa.ForeignKeyConstraint(['model_id'], ['model_metadata.models.model_id'], ),
    sa.PrimaryKeyConstraint('model_id', 'subset_hash', 'tie_breaker', 'evaluation_start_time', 'evaluation_end_time', 'parameter', 'attribute_name', 'attribute_value'),
    schema='train_results'
    )
    op.create_index(op.f('ix_train_results_aequitas_attribute_name'), 'aequitas', ['attribute_name'], unique=False, schema='train_results')
    op.create_index(op.f('ix_train_results_aequitas_attribute_value'), 'aequitas', ['attribute_value'], unique=False, schema='train_results')
    op.create_index(op.f('ix_train_results_aequitas_evaluation_end_time'), 'aequitas', ['evaluation_end_time'], unique=False, schema='train_results')
    op.create_index(op.f('ix_train_results_aequitas_evaluation_start_time'), 'aequitas', ['evaluation_start_time'], unique=False, schema='train_results')
    op.create_index(op.f('ix_train_results_aequitas_model_id'), 'aequitas', ['model_id'], unique=False, schema='train_results')
    op.create_index(op.f('ix_train_results_aequitas_parameter'), 'aequitas', ['parameter'], unique=False, schema='train_results')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_train_results_aequitas_parameter'), table_name='aequitas', schema='train_results')
    op.drop_index(op.f('ix_train_results_aequitas_model_id'), table_name='aequitas', schema='train_results')
    op.drop_index(op.f('ix_train_results_aequitas_evaluation_start_time'), table_name='aequitas', schema='train_results')
    op.drop_index(op.f('ix_train_results_aequitas_evaluation_end_time'), table_name='aequitas', schema='train_results')
    op.drop_index(op.f('ix_train_results_aequitas_attribute_value'), table_name='aequitas', schema='train_results')
    op.drop_index(op.f('ix_train_results_aequitas_attribute_name'), table_name='aequitas', schema='train_results')
    op.drop_table('aequitas', schema='train_results')
    op.drop_index(op.f('ix_test_results_aequitas_parameter'), table_name='aequitas', schema='test_results')
    op.drop_index(op.f('ix_test_results_aequitas_model_id'), table_name='aequitas', schema='test_results')
    op.drop_index(op.f('ix_test_results_aequitas_evaluation_start_time'), table_name='aequitas', schema='test_results')
    op.drop_index(op.f('ix_test_results_aequitas_evaluation_end_time'), table_name='aequitas', schema='test_results')
    op.drop_index(op.f('ix_test_results_aequitas_attribute_value'), table_name='aequitas', schema='test_results')
    op.drop_index(op.f('ix_test_results_aequitas_attribute_name'), table_name='aequitas', schema='test_results')
    op.drop_table('aequitas', schema='test_results')
    # ### end Alembic commands ###



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import os
import numpy as np
import pandas as pd

from .utils import str_in_sql
from .metric_directionality import sql_rank_order, value_agg_funcs
from .plotting import plot_cats, plot_bounds, category_colordict, category_styledict


class DistanceFromBestTable:
    def __init__(self, db_engine, models_table, distance_table, agg_type):
        """A database table that stores the distance from models and the
        best model for that train end time for a variety of chosen metrics

        Args:
            db_engine (sqlalchemy.engine)
            models_table (string) The name of a models table in the database, pre-populated
            distance_table (string) The desired name of the distance table to be
                produced by this class
            agg_type (string) Method for aggregating metric values (for instance, if there
                are multiple models at a given train_end_time with different random seeds)
        """
        self.db_engine = db_engine
        self.models_table = models_table
        self.distance_table = distance_table
        self.agg_type = agg_type

    def _delete(self):
        """Delete the distance-from-best table if it exists"""
        self.db_engine.execute("drop table if exists {}".format(self.distance_table))

    def _create(self):
        """Create the distance-from-best table"""
        self.db_engine.execute(
            """create table {} (
            model_group_id int,
            train_end_time timestamp,
            metric text,
            parameter text,
            raw_value float,
            best_case float,
            dist_from_best_case float,
            raw_value_next_time float,
            dist_from_best_case_next_time float
        )""".format(
                self.distance_table
            )
        )

    def _populate(self, model_group_ids, train_end_times, metrics):
        """Populate the distance table with the given model groups, times, and metrics

        Args:
            model_group_ids (list) Model group ids to include in the distance table
            train_end_times (list) Train end times to include in the table
            metrics (list) Metrics and metric params to include in the table. Each
                row should be a dict with keys:
                        'metric' (e.g. 'precision@')
                        'parameter' (e.g. '100_abs')
                All models should have the test_results.evaluations table populated
                for all given model group ids, train end times, and metric/param combos
        """
        logger.debug("Populating data to distance table")
        for metric in metrics:
            self.db_engine.execute(
                """
                insert into {new_table}
                WITH first_evals AS (
                    SELECT *, row_number() OVER (
                        PARTITION BY model_id
                        ORDER BY evaluation_start_time ASC, evaluation_end_time ASC
                        ) AS eval_rn
                    FROM test_results.evaluations
                    WHERE metric='{metric}' AND parameter='{parameter}' AND subset_hash=''
                ),
                metric_values AS (
                    SELECT
                        m.model_group_id,
                        m.train_end_time,
                        {metric_agg_fcn}(ev.stochastic_value) as value
                  FROM first_evals ev
                  JOIN triage_metadata.{models_table} m USING(model_id)
                  JOIN triage_metadata.model_groups mg USING(model_group_id)
                  WHERE m.model_group_id IN ({model_group_ids})
                        AND train_end_time in ({train_end_times})
                        AND ev.eval_rn = 1
                  GROUP BY model_group_id, train_end_time
                ),
                model_ranks AS (
                    SELECT
                        model_group_id,
                        train_end_time,
                        value,
                        row_number() OVER (
                            PARTITION BY train_end_time
                            ORDER BY value {metric_value_order}, RANDOM()
                        ) AS rank
                  FROM metric_values
                ),
                model_tols AS (
                  SELECT train_end_time, model_group_id,
                         rank,
                         value,
                         first_value(value) over (
                            partition by train_end_time
                            order by rank ASC
                        ) AS best_val
                  FROM model_ranks
                ),
                current_best_vals as (
                    SELECT
                        model_group_id,
                        train_end_time,
                        '{metric}',
                        '{parameter}',
                        value as raw_value,
                        best_val as best_case,
                        abs(value - best_val) dist_from_best_case
                    FROM model_tols
                )
                select
                    current_best_vals.*,
                    first_value(raw_value) over (
                        partition by model_group_id
                        order by train_end_time asc
                        rows between 1 following and unbounded following
                    ) raw_value_next_time,
                    first_value(dist_from_best_case) over (
                        partition by model_group_id
                        order by train_end_time asc
                        rows between 1 following and unbounded following
                    ) dist_from_best_case_next_time
                from current_best_vals
                order by train_end_time
            """.format(
                    model_group_ids=str_in_sql(model_group_ids),
                    train_end_times=str_in_sql(train_end_times),
                    models_table=self.models_table,
                    metric=metric["metric"],
                    parameter=metric["parameter"],
                    metric_value_order=sql_rank_order(metric["metric"]),
                    new_table=self.distance_table,
                    metric_agg_fcn=value_agg_funcs(metric["metric"])[self.agg_type],
                )
            )

    @property
    def observed_bounds(self):
        query = """
            SELECT
                metric,
                parameter,
                min(raw_value),
                max(raw_value)
            FROM {distance_table} dist
            GROUP BY metric, parameter
        """.format(
            distance_table=self.distance_table
        )
        return dict(
            ((metric, parameter), (minimum, maximum))
            for metric, parameter, minimum, maximum in self.db_engine.execute(query)
        )

    def create_and_populate(
        self, model_group_ids, train_end_times, metrics, delete=True
    ):
        """Creates and populates the distance table with the
            given model groups, times, and metrics

        Args:
            model_group_ids (list) Model group ids to include in the distance table
            train_end_times (list) Train end times to include in the table
            metrics (list) Metrics and parameters to include in the table. Each
                row should be a dict with keys:
                        'metric' (e.g. 'precision@')
                        'parameter' (e.g. '100_abs')
                All models should have the test_results.evaluations table populated
                for all given model group ids, train end times, and metric/param combos
            delete (boolean, optional) Delete any previous version of the
                distance table if it exists
        """
        if delete:
            self._delete()
        self._create()
        self._populate(model_group_ids, train_end_times, metrics)

    def as_dataframe(self, model_group_ids):
        """Return model-group-id subset of table as dataframe

        Args:
            model_group_ids (list) the desired model group ids

        Returns: (pandas.DataFrame) The data from the table corresponding
            to those model group ids
        """
        return pd.read_sql(
            "select * from {} where model_group_id in ({})".format(
                self.distance_table, str_in_sql(model_group_ids)
            ),
            self.db_engine,
        )

    def dataframe_as_of(self, model_group_ids, train_end_time):
        """Return model group id/train end time subset of table as dataframe

        Args:
            model_group_ids (list) the desired model group ids
            train_end_date (string) the desired train end time

        Returns: (pandas.DataFrame) The data from the table corresponding
            to those model group ids and train end time
        """
        base_df = self.as_dataframe(model_group_ids)
        return base_df[base_df["train_end_time"] == train_end_time]


class BestDistancePlotter:
    def __init__(self, distance_from_best_table, directory=None):
        """Generate a plot illustrating the effect of different below-best maximum
        thresholds across the dataset.

        Args:
            distance_from_best_table (audition.DistanceFromBestTable)
                A pre-populated distance-from-best database table
        """
        self.distance_from_best_table = distance_from_best_table
        self.directory = directory

        self.colordict = None
        self.styledict = None
        self.cmap_name = "tab10"

    def plot_bounds(self, metric, parameter):
        observed_min, observed_max = self.distance_from_best_table.observed_bounds[
            (metric, parameter)
        ]
        return plot_bounds(observed_min, observed_max)

    def plot_tick_dist(self, plot_min, plot_max):
        dist = plot_max - plot_min
        return dist / 100.0

    def generate_plot_data(self, metric, parameter, model_group_ids, train_end_times):
        """Fetch data necessary for producing the plot from the distance table

        Arguments:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter,
                such as '300_abs'
            model_group_ids (list) - Model group ids to include in the dataset
            train_end_times (list) - Train end times to include in the dataset

        Returns: (pandas.DataFrame) The relevant models and the percentage of time
            each was within various thresholds of the best model at that time
        """
        model_group_union_sql = " union all ".join(
            [
                "(select {} as model_group_id)".format(model_group_id)
                for model_group_id in model_group_ids
            ]
        )
        plot_min, plot_max = self.plot_bounds(metric, parameter)
        plot_tick_dist = self.plot_tick_dist(plot_min, plot_max)
        sel_params = {
            "metric": metric,
            "parameter": parameter,
            "model_group_union_sql": model_group_union_sql,
            "distance_table": self.distance_from_best_table.distance_table,
            "model_group_str": str_in_sql(model_group_ids),
            "train_end_str": str_in_sql(train_end_times),
            "series_start": plot_min,
            "series_end": plot_max,
            "series_tick": plot_tick_dist,
        }
        sel = """\
            with model_group_ids as ({model_group_union_sql}),
            x_vals AS (
                SELECT m.model_group_id, s.distance
                FROM (SELECT GENERATE_SERIES(
                {series_start}, {series_end}, {series_tick}
                ) AS distance) s
                CROSS JOIN
                (
                SELECT DISTINCT model_group_id FROM model_group_ids
                ) m
            )
            SELECT dist.model_group_id, distance, mg.model_type,
                    COUNT(*) AS num_models,
                    AVG(CASE WHEN dist_from_best_case <= distance THEN 1 ELSE 0 END) AS pct_of_time
            FROM {distance_table} dist
            JOIN x_vals USING(model_group_id)
            JOIN triage_metadata.model_groups mg using (model_group_id)
            WHERE
                dist.metric='{metric}'
                AND dist.parameter='{parameter}'
                and model_group_id in ({model_group_str})
                and train_end_time in ({train_end_str})
            GROUP BY 1,2,3
        """.format(
            **sel_params
        )

        return pd.read_sql(sel, self.distance_from_best_table.db_engine).sort_values(
            ["model_group_id", "distance"]
        )

    def plot_all_best_dist(self, metric_filters, model_group_ids, train_end_times):
        """For each metric, plot the percentage of time that a model group is
        within X percentage points of the best-performing model group using that
        metric.

        Arguments:
            metric_filters (list) The metrics to plot. Each element should be
                a dict with the following keys:

                metric (string) -- model evaluation metric, such as 'precision@'
                parameter (string) -- model evaluation metric parameter,
                    such as '300_abs'
            model_group_ids (list) - Model group ids to include in the plot
            train_end_times (list) - Train end times to include in the plot

        """
        for metric_filter in metric_filters:
            logger.debug(
                "Building best distance plot for %s and %s",
                metric_filter,
                train_end_times,
            )
            df = self.generate_plot_data(
                metric=metric_filter["metric"],
                parameter=metric_filter["parameter"],
                model_group_ids=model_group_ids,
                train_end_times=train_end_times,
            )

            # set stable colors/styles by model type
            categories = np.unique(df['model_type'])
            if not self.colordict:
                self.colordict = category_colordict(self.cmap_name, categories, None)
            if not self.styledict:
                self.styledict = category_styledict(self.colordict, None)

            plot_best_dist(
                metric=metric_filter["metric"],
                parameter=metric_filter["parameter"],
                df_best_dist=df,
                directory=self.directory,
                colordict=self.colordict,
                styledict=self.styledict,
            )


def plot_best_dist(metric, parameter, df_best_dist, directory=None, **plt_format_args):
    """Generates a plot of the percentage of time that a model group is
    within X percentage points of the best-performing model group using a
    given metric. At each point in time that a set of model groups is
    evaluated, the performance of the best model is calculated and the
    difference in performace for all other models found relative to this.

    An (x,y) point for a given model group on the plot generated by this
    method means that across all of those tets sets, the model
    from that model group performed within X percentage points of the best
    model in y% of the test sets.

    The plot will contain a line for each given model group representing
    the cumulative percent of the time that the group is within Xpp of the
    best group for each value of X between 0 and 100. All groups ultimately
    reach (1,1) on this graph (as every model group must be within 100pp of
    the best model 100% of the time), and a model specification that always
    dominated the others in the experiment would start at (0,1) and remain
    at y=1 across the graph.

    Arguments:
        metric (string) -- model evaluation metric, such as 'precision@'
        parameter (string) -- model evaluation metric parameter, such as '300_abs'
        df_best_dist (pandas.DataFrame)
        **plt_format_args -- formatting arguments passed through to plot_cats()
    """

    cat_col = "model_type"
    plt_title = "Fraction of models X pp worse than best {} {}".format(
        metric, parameter
    )

    if directory:
        path_to_save = os.path.join(
            directory, f"distance_from_best_{metric}{parameter}.png"
        )
    else:
        path_to_save = None

    plot_cats(
        frame=df_best_dist,
        x_col="distance",
        y_col="pct_of_time",
        cat_col=cat_col,
        title=plt_title,
        x_label="distance from best {}".format(metric),
        y_label="fraction of models",
        x_ticks=np.arange(0, 1.1, 0.1),
        path_to_save=path_to_save,
        **plt_format_args,
    )



"""rankify_predictions

Revision ID: 609c7cc51794
Revises: 50e1f1bc2cac
Create Date: 2019-04-03 16:23:22.903979

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '609c7cc51794'
down_revision = '97cf99b7348f'
branch_labels = None
depends_on = None


def add_rank_abs_with_ties_to_predictions(schema, op):
    """Add the new column rank_abs_with_ties directly after rank_abs

    Altering a table to add a column in Postgres can only put columns at the end of the table.
    This is not ideal. So we recreate the table and copy everything over.
    """
    op.execute(f"alter table {schema}.predictions rename to predictions_old")
    op.create_table(
        "predictions",
        sa.Column("model_id", sa.Integer(), nullable=False),
        sa.Column("entity_id", sa.BigInteger(), nullable=False),
        sa.Column("as_of_date", sa.DateTime(), nullable=False),
        sa.Column("score", sa.Numeric(6, 5), nullable=True),
        sa.Column("label_value", sa.Integer(), nullable=True),
        sa.Column("rank_abs_no_ties", sa.Integer(), nullable=True),
        sa.Column("rank_abs_with_ties", sa.Integer(), nullable=True),
        sa.Column("rank_pct_no_ties", sa.Numeric(6, 5), nullable=True),
        sa.Column("rank_pct_with_ties", sa.Numeric(6, 5), nullable=True),
        sa.Column("matrix_uuid", sa.Text(), nullable=True),
        sa.Column("test_label_timespan", sa.Interval(), nullable=True),
        schema=schema,
    )
    op.execute(f"""
insert into {schema}.predictions
select model_id, entity_id, as_of_date, score, label_value, rank_abs, null, rank_pct, null, matrix_uuid, test_label_timespan from {schema}.predictions_old
""")
    op.drop_table("predictions_old", schema=schema)
    op.create_primary_key(
        f"pk_{schema}_predictions",
        "predictions",
        schema=schema,
        columns=['model_id', 'entity_id', 'as_of_date']
    )
    op.create_foreign_key(
        constraint_name=f"{schema}_predictions_matrix_uuid_fkey",
        source_table="predictions",
        source_schema=schema,
        referent_table="matrices",
        referent_schema="model_metadata",
        local_cols=["matrix_uuid"],
        remote_cols=["matrix_uuid"],
    )
    op.create_foreign_key(
        constraint_name=f"{schema}_predictions_model_id_fkey",
        source_table="predictions",
        source_schema=schema,
        referent_table="models",
        referent_schema="model_metadata",
        local_cols=["model_id"],
        remote_cols=["model_id"],
    )

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('prediction_metadata',
    sa.Column('model_id', sa.Integer(), nullable=False),
    sa.Column('matrix_uuid', sa.Text(), nullable=False),
    sa.Column('tiebreaker_ordering', sa.Text(), nullable=True),
    sa.Column('random_seed', sa.Integer(), nullable=True),
    sa.Column('predictions_saved', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['matrix_uuid'], ['model_metadata.matrices.matrix_uuid'], ),
    sa.ForeignKeyConstraint(['model_id'], ['model_metadata.models.model_id'], ),
    sa.PrimaryKeyConstraint('model_id', 'matrix_uuid'),
    schema='test_results'
    )
    op.create_table('prediction_metadata',
    sa.Column('model_id', sa.Integer(), nullable=False),
    sa.Column('matrix_uuid', sa.Text(), nullable=False),
    sa.Column('tiebreaker_ordering', sa.Text(), nullable=True),
    sa.Column('random_seed', sa.Integer(), nullable=True),
    sa.Column('predictions_saved', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['matrix_uuid'], ['model_metadata.matrices.matrix_uuid'], ),
    sa.ForeignKeyConstraint(['model_id'], ['model_metadata.models.model_id'], ),
    sa.PrimaryKeyConstraint('model_id', 'matrix_uuid'),
    schema='train_results'
    )
    add_rank_abs_with_ties_to_predictions("train_results", op)
    add_rank_abs_with_ties_to_predictions("test_results", op)
    op.add_column('models', sa.Column('random_seed', sa.Integer(), nullable=True), schema='model_metadata')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('models', 'random_seed', schema='model_metadata')
    op.drop_column('predictions', 'rank_abs_with_ties', schema='train_results')
    op.drop_column('predictions', 'rank_abs_with_ties', schema='test_results')
    op.drop_column('predictions', 'rank_pct_with_ties', schema='train_results')
    op.drop_column('predictions', 'rank_pct_with_ties', schema='test_results')
    op.alter_column('predictions', 'rank_abs_no_ties', new_column_name='rank_abs', schema='train_results')
    op.alter_column('predictions', 'rank_pct_no_ties', new_column_name='rank_pct', schema='train_results')
    op.alter_column('predictions', 'rank_abs_no_ties', new_column_name='rank_abs', schema='test_results')
    op.alter_column('predictions', 'rank_pct_no_ties', new_column_name='rank_pct', schema='test_results')
    op.drop_table('prediction_metadata', schema='train_results')
    op.drop_table('prediction_metadata', schema='test_results')
    # ### end Alembic commands ###



from datetime import datetime
import tempfile
import os
import mock

import factory
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition import Auditioner, AuditionRunner
from triage.component.catwalk.db import ensure_db

from tests.results_tests.factories import (
    EvaluationFactory,
    ModelFactory,
    ModelGroupFactory,
    init_engine,
    session,
)

config = {
    "filter": {
        "distance_table": "distance_table",
        "max_from_best": 1.0,
        "metric": "precision@",
        "models_table": "models",
        "parameter": "50_abs",
        "threshold_value": 0.0,
    },
    "model_groups": {
        "query": "SELECT DISTINCT(model_group_id) FROM triage_metadata.model_groups"
    },
    "rules": [
        {
            "selection_rules": [
                {"name": "best_current_value", "n": 3},
                {"name": "best_average_value", "n": 3},
                {"name": "lowest_metric_variance", "n": 3},
                {
                    "dist_from_best_case": [0.05],
                    "name": "most_frequent_best_dist",
                    "n": 3,
                },
            ],
            "shared_parameters": [{"metric": "precision@", "parameter": "50_abs"}],
        },
        {
            "selection_rules": [
                {
                    "metric1_weight": 0.5,
                    "metric2": "recall@",
                    "name": "best_average_two_metrics",
                    "n": 3,
                    "parameter2": "50_abs",
                }
            ],
            "shared_parameters": [{"metric1": "precision@", "parameter1": "50_abs"}],
        },
    ],
    "time_stamps": {
        "query": "SELECT DISTINCT train_end_time FROM triage_metadata.models "
        "WHERE model_group_id IN ({}) AND EXTRACT(DAY FROM train_end_time) IN (1) "
        "AND train_end_time >= '2012-01-01'"
    },
}


def test_Audition():
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        init_engine(db_engine)

        num_model_groups = 10
        model_types = [
            "classifier type {}".format(i) for i in range(0, num_model_groups)
        ]
        model_groups = [
            ModelGroupFactory(model_type=model_type) for model_type in model_types
        ]
        train_end_times = [
            datetime(2013, 1, 1),
            datetime(2014, 1, 1),
            datetime(2015, 1, 1),
            datetime(2016, 1, 1),
        ]

        models = [
            ModelFactory(model_group_rel=model_group, train_end_time=train_end_time)
            for model_group in model_groups
            for train_end_time in train_end_times
        ]
        metrics = [
            ("precision@", "100_abs"),
            ("recall@", "100_abs"),
            ("precision@", "50_abs"),
            ("recall@", "50_abs"),
            ("fpr@", "10_pct"),
        ]

        class ImmediateEvalFactory(EvaluationFactory):
            evaluation_start_time = factory.LazyAttribute(
                lambda o: o.model_rel.train_end_time
            )

        for model in models:
            for (metric, parameter) in metrics:
                ImmediateEvalFactory(
                    model_rel=model, metric=metric, parameter=parameter
                )

        session.commit()

        with tempfile.TemporaryDirectory() as td:
            with mock.patch('os.getcwd') as mock_getcwd:
                mock_getcwd.return_value = td
                AuditionRunner(config_dict=config, db_engine=db_engine, directory=td).run()
                assert len(os.listdir(os.getcwd())) == 6


def test_Auditioner():
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        init_engine(db_engine)
        # set up data, randomly generated by the factories but conforming
        # generally to what we expect triage_metadata schema data to look like

        num_model_groups = 10
        model_types = [
            "classifier type {}".format(i) for i in range(0, num_model_groups)
        ]
        model_groups = [
            ModelGroupFactory(model_type=model_type) for model_type in model_types
        ]
        train_end_times = [
            datetime(2013, 1, 1),
            datetime(2014, 1, 1),
            datetime(2015, 1, 1),
            datetime(2016, 1, 1),
        ]

        models = [
            ModelFactory(model_group_rel=model_group, train_end_time=train_end_time)
            for model_group in model_groups
            for train_end_time in train_end_times
        ]
        metrics = [
            ("precision@", "100_abs"),
            ("recall@", "100_abs"),
            ("precision@", "50_abs"),
            ("recall@", "50_abs"),
            ("fpr@", "10_pct"),
        ]

        class ImmediateEvalFactory(EvaluationFactory):
            evaluation_start_time = factory.LazyAttribute(
                lambda o: o.model_rel.train_end_time
            )

        for model in models:
            for (metric, parameter) in metrics:
                ImmediateEvalFactory(
                    model_rel=model, metric=metric, parameter=parameter
                )

        session.commit()

        # define a very loose filtering that should admit all model groups
        no_filtering = [
            {
                "metric": "precision@",
                "parameter": "100_abs",
                "max_from_best": 1.0,
                "threshold_value": 0.0,
            },
            {
                "metric": "recall@",
                "parameter": "100_abs",
                "max_from_best": 1.0,
                "threshold_value": 0.0,
            },
        ]
        model_group_ids = [mg.model_group_id for mg in model_groups]
        auditioner = Auditioner(
            db_engine, model_group_ids, train_end_times, no_filtering
        )
        assert len(auditioner.thresholded_model_group_ids) == num_model_groups
        auditioner.plot_model_groups()

        # here, we pick thresholding rules that should definitely remove
        # all model groups from contention because they are too strict.
        remove_all = [
            {
                "metric": "precision@",
                "parameter": "100_abs",
                "max_from_best": 0.0,
                "threshold_value": 1.1,
            },
            {
                "metric": "recall@",
                "parameter": "100_abs",
                "max_from_best": 0.0,
                "threshold_value": 1.1,
            },
        ]

        auditioner.update_metric_filters(new_filters=remove_all)
        assert len(auditioner.thresholded_model_group_ids) == 0

        # pass the argument instead and remove all model groups
        auditioner.set_one_metric_filter(
            metric="precision@",
            parameter="100_abs",
            max_from_best=0.0,
            threshold_value=1.1,
        )
        assert len(auditioner.thresholded_model_group_ids) == 0

        # one potential place for bugs would be when we pull back the rules
        # for being too restrictive. we want to make sure that the original list is
        # always used for thresholding, or else such a move would be impossible
        auditioner.update_metric_filters(new_filters=no_filtering)
        assert len(auditioner.thresholded_model_group_ids) == num_model_groups

        # pass the argument instead and let all model groups pass
        auditioner.set_one_metric_filter(
            metric="precision@",
            parameter="100_abs",
            max_from_best=1.0,
            threshold_value=0.0,
        )
        assert len(auditioner.thresholded_model_group_ids) == num_model_groups

        # now, we want to take this partially thresholded list and run it through
        # a grid of selection rules, meant to pick winners by a variety of user-defined
        # criteria
        rule_grid = [
            {
                "shared_parameters": [
                    {"metric": "precision@", "parameter": "100_abs"},
                    {"metric": "recall@", "parameter": "100_abs"},
                ],
                "selection_rules": [
                    {
                        "name": "most_frequent_best_dist",
                        "dist_from_best_case": [0.1, 0.2, 0.3],
                        "n": 1,
                    },
                    {"name": "best_current_value", "n": 1},
                ],
            },
            {
                "shared_parameters": [
                    {"metric1": "precision@", "parameter1": "100_abs"}
                ],
                "selection_rules": [
                    {
                        "name": "best_average_two_metrics",
                        "metric2": ["recall@"],
                        "parameter2": ["100_abs"],
                        "metric1_weight": [0.4, 0.5, 0.6],
                        "n": 1,
                    }
                ],
            },
        ]
        auditioner.register_selection_rule_grid(rule_grid, plot=False)
        final_model_group_ids = auditioner.selection_rule_model_group_ids

        # we expect the result to be a mapping of selection rule name to model group id
        assert isinstance(final_model_group_ids, dict)

        # we expect that there is one winner for each selection rule
        assert sorted(final_model_group_ids.keys()) == sorted(
            [rule.descriptive_name for rule in auditioner.selection_rules]
        )



"""empty message

Revision ID: 264245ddfce2
Revises: 0d44655e35fd
Create Date: 2017-09-01 14:26:01.107455

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "264245ddfce2"
down_revision = "0d44655e35fd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "individual_importances",
        sa.Column("feature_value", sa.Float(), nullable=True),
        schema="results",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("individual_importances", "feature_value", schema="results")
    # ### end Alembic commands ###



from contextlib import contextmanager
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import make_transient
import datetime
from unittest.mock import Mock
from numpy.testing import assert_array_almost_equal
import pandas as pd

from triage.component.results_schema import TestPrediction, Matrix, Model
from triage.component.catwalk.storage import TestMatrixType
from triage.component.catwalk.db import ensure_db
from tests.results_tests.factories import (
    MatrixFactory,
    ModelFactory,
    PredictionFactory,
    init_engine,
    session as factory_session
)
from triage.database_reflection import table_has_data

from triage.component.catwalk.predictors import Predictor
from tests.utils import (
    MockTrainedModel,
    matrix_creator,
    matrix_metadata_creator,
    get_matrix_store,
    rig_engines,
)
import pytest


with_matrix_types = pytest.mark.parametrize(
    ('matrix_type',),
    [
        ('train',),
        ('test',),
    ],
)

MODEL_RANDOM_SEED = 123456


@contextmanager
def prepare():
    with rig_engines() as (db_engine, project_storage):
        train_matrix_uuid = "1234"
        try:
            session = sessionmaker(db_engine)()
            session.add(Matrix(matrix_uuid=train_matrix_uuid))

            # Create the fake trained model and store in db
            trained_model = MockTrainedModel()
            model_hash = "abcd"
            project_storage.model_storage_engine().write(trained_model, model_hash)
            db_model = Model(
                model_hash=model_hash,
                train_matrix_uuid=train_matrix_uuid,
                random_seed=MODEL_RANDOM_SEED
            )
            session.add(db_model)
            session.commit()
            yield project_storage, db_engine, db_model.model_id
        finally:
            session.close()


@pytest.fixture(name='predict_setup_args', scope='function')
def fixture_predict_setup_args():
    with prepare() as predict_setup_args:
        yield predict_setup_args


@pytest.fixture(name='predictor', scope='function')
def predictor(predict_setup_args):
    (project_storage, db_engine, model_id) = predict_setup_args
    return Predictor(project_storage.model_storage_engine(), db_engine, rank_order='worst')


@pytest.fixture(name='predict_proba', scope='function')
def prediction_results(matrix_type, predictor, predict_setup_args):
    (project_storage, db_engine, model_id) = predict_setup_args

    dayone = datetime.datetime(2011, 1, 1)
    daytwo = datetime.datetime(2011, 1, 2)
    source_dict = {
        "entity_id": [1, 2, 3, 1, 2, 3],
        "as_of_date": [dayone, dayone, dayone, daytwo, daytwo, daytwo],
        "feature_one": [3] * 6,
        "feature_two": [5] * 6,
        "label": [True, False] * 3
    }

    matrix = pd.DataFrame.from_dict(source_dict)
    metadata = matrix_metadata_creator(matrix_type=matrix_type)
    matrix_store = get_matrix_store(project_storage, matrix, metadata)

    predict_proba = predictor.predict(
        model_id,
        matrix_store,
        misc_db_parameters=dict(),
        train_matrix_columns=["feature_one", "feature_two"],
    )
    return predict_proba


@with_matrix_types
def test_predictor(predict_proba):
    """assert that the returned predictions are of the desired length"""
    assert len(predict_proba) == 6


@with_matrix_types
def test_predictions_table(predictor, predict_proba, matrix_type):
    """assert that the predictions table entries are present, linked to the original models"""
    records = [
        row
        for row in predictor.db_engine.execute(
            """select entity_id, as_of_date
        from {}_results.predictions
        join triage_metadata.models using (model_id)""".format(
                matrix_type, matrix_type
            )
        )
    ]
    assert len(records) == 6





@with_matrix_types
def test_predictor_save_predictions(matrix_type, predict_setup_args):
    """Test the save_predictions flag being set to False

    We still want to return predict_proba, but not save data to the DB
    """
    (project_storage, db_engine, model_id) = predict_setup_args
    # if save_predictions is sent as False, don't save
    predictor = Predictor(project_storage.model_storage_engine(), db_engine, rank_order='worst', save_predictions=False)

    matrix_store = get_matrix_store(project_storage)
    train_matrix_columns = matrix_store.columns()

    predict_proba = predictor.predict(
        model_id,
        matrix_store,
        misc_db_parameters=dict(),
        train_matrix_columns=train_matrix_columns,
    )

    # assert
    # 1. that the returned predictions are of the desired length
    assert len(predict_proba) == 2

    # 2. that the predictions table entries are present and
    # can be linked to the original models
    assert not table_has_data(f"{matrix_type}_predictions", db_engine)


@with_matrix_types
def test_predictor_needs_predictions(matrix_type, predict_setup_args):
    """Test that the logic that figures out if predictions are needed for a given model/matrix"""
    (project_storage, db_engine, model_id) = predict_setup_args
    # if not all of the predictions for the given model id and matrix are present in the db,
    # needs_predictions should return true. else, false
    predictor = Predictor(project_storage.model_storage_engine(), db_engine, 'worst')

    metadata = matrix_metadata_creator(matrix_type=matrix_type)
    matrix_store = get_matrix_store(project_storage, metadata=metadata)
    train_matrix_columns = matrix_store.columns()

    # we haven't done anything yet, this should definitely need predictions
    assert predictor.needs_predictions(matrix_store, model_id)
    predictor.predict(
        model_id,
        matrix_store,
        misc_db_parameters=dict(),
        train_matrix_columns=train_matrix_columns,
    )
    # now that predictions have been made, this should no longer need predictions
    assert not predictor.needs_predictions(matrix_store, model_id)


def test_predictor_get_train_columns(predict_setup_args):
    """Test behavior when train/test matrices are created with different column orders
    """
    (project_storage, db_engine, model_id) = predict_setup_args
    predictor = Predictor(project_storage.model_storage_engine(), db_engine, 'worst')
    train_store = get_matrix_store(
        project_storage=project_storage,
        matrix=matrix_creator(),
        metadata=matrix_metadata_creator(matrix_type="train"),
    )

    # flip the order of some feature columns in the test matrix
    other_order_matrix = matrix_creator()
    order = other_order_matrix.columns.tolist()
    order[0], order[1] = order[1], order[0]
    other_order_matrix = other_order_matrix[order]
    test_store = get_matrix_store(
        project_storage=project_storage,
        matrix=other_order_matrix,
        metadata=matrix_metadata_creator(matrix_type="test"),
    )

    # Runs the same test for training and testing predictions
    for store, mat_type in zip((train_store, test_store), ("train", "test")):
        predict_proba = predictor.predict(
            model_id,
            store,
            misc_db_parameters=dict(),
            train_matrix_columns=train_store.columns(),
        )
        # assert
        # 1. that we calculated predictions
        assert len(predict_proba) > 0

        # 2. that the predictions table entries are present and
        # can be linked to the original models
        records = [
            row
            for row in db_engine.execute(
                """select entity_id, as_of_date
            from {}_results.predictions
            join triage_metadata.models using (model_id)""".format(
                    mat_type, mat_type
                )
            )
        ]
        assert len(records) > 0


def test_predictor_retrieve(predict_setup_args):
    """Test the predictions retrieved from the database match the output from predict_proba"""
    (project_storage, db_engine, model_id) = predict_setup_args
    predictor = Predictor(
        project_storage.model_storage_engine(), db_engine, 'worst', replace=False
    )

    # create prediction set
    matrix_store = get_matrix_store(project_storage)

    predict_proba = predictor.predict(
        model_id,
        matrix_store,
        misc_db_parameters=dict(),
        train_matrix_columns=matrix_store.columns()
    )

    # When run again, the predictions retrieved from the database
    # should match.
    #
    # Some trickiness here. Let's explain:
    #
    # If we are not careful, retrieving predictions from the database and
    # presenting them as a numpy array can result in a bad ordering,
    # since the given matrix may not be 'ordered' by some criteria
    # that can be easily represented by an ORDER BY clause.
    #
    # It will sometimes work, because without ORDER BY you will get
    # it back in the table's physical order, which unless something has
    # happened to the table will be the order you inserted it,
    # which could very well be the order in the matrix.
    # So it's not a bug that would necessarily immediately show itself,
    # but when it does go wrong your scores will be garbage.
    #
    # So we simulate a table order mutation that can happen over time:
    # Remove the first row and put it at the end.
    # If the Predictor doesn't explicitly reorder the results, this will fail
    # Only running on TestPrediction because TrainPrediction behaves the exact same way
    try:
        reorder_session = sessionmaker(bind=db_engine)()
        obj = reorder_session.query(TestPrediction).first()
        reorder_session.delete(obj)
        reorder_session.commit()
    finally:
        reorder_session.close()

    make_transient(obj)
    try:
        reorder_session = sessionmaker(bind=db_engine)()
        reorder_session.add(obj)
        reorder_session.commit()
    finally:
        reorder_session.close()

    predictor.load_model = Mock()
    new_predict_proba = predictor.predict(
        model_id,
        matrix_store,
        misc_db_parameters=dict(),
        train_matrix_columns=matrix_store.columns()
    )
    assert_array_almost_equal(new_predict_proba, predict_proba, decimal=5)
    assert not predictor.load_model.called



from triage.component.catwalk.evaluation import (
    SORT_TRIALS,
    ModelEvaluator,
    generate_binary_at_x,
    query_subset_table,
    subset_labels_and_predictions,
)
from triage.component.catwalk.metrics import Metric
import testing.postgresql
import datetime
import re

import factory
import numpy as np
from numpy.testing import assert_almost_equal, assert_array_equal
import pandas as pd
from sqlalchemy.sql.expression import text
from triage.component.catwalk.utils import filename_friendly_hash, get_subset_table_name
from triage.component.catwalk.storage import MatrixStore
from tests.utils import fake_labels, fake_trained_model, MockMatrixStore
from tests.results_tests.factories import (
    ModelFactory,
    EvaluationFactory,
    PredictionFactory,
    SubsetFactory,
    session,
)


@Metric(greater_is_better=True)
def always_half(predictions_proba, predictions_binary, labels, parameters):
    return 0.5


SUBSETS = [
    {
        "name": "evens",
        "query": """
            select distinct entity_id
            from events
            where entity_id % 2 = 0
            and outcome_date < '{as_of_date}'::date
        """,
    },
    {
        "name": "odds",
        "query": """
            select distinct entity_id
            from events
            where entity_id % 2 = 1
            and outcome_date < '{as_of_date}'::date
        """,
    },
    {
        "name": "empty",
        "query": """
            select distinct entity_id
            from events
            where entity_id = -1
            and outcome_date < '{as_of_date}'::date
        """,
    },
]

TRAIN_END_TIME = datetime.datetime(2016, 1, 1)


def populate_subset_data(db_engine, subset, entity_ids, as_of_date=TRAIN_END_TIME):
    table_name = get_subset_table_name(subset)
    query_where_clause = re.search("where.*[0-9]", subset["query"]).group()

    db_engine.execute(
        f"""
        create table {table_name} (
            entity_id int,
            as_of_date date,
            active bool
        )
        """
    )

    for entity_id in entity_ids:
        insert_query = f"""
            with unfiltered_row as (
                select {entity_id} as entity_id,
                       '{as_of_date}'::date as as_of_date,
                       true as active
            )
            insert into {table_name}
            select entity_id, as_of_date, active
            from unfiltered_row
            {query_where_clause}
            """
        db_engine.execute(
            text(insert_query).execution_options(autocommit=True))


def test_all_same_labels(db_engine_with_results_schema):
    num_entities = 5
    trained_model, model_id = fake_trained_model(
        db_engine_with_results_schema,
        train_end_time=TRAIN_END_TIME,
    )

    for label_value in [0, 1]:
        labels = [label_value] * num_entities

        # We should be able to calculate accuracy even if all of the labels
        # are the same, but ROC_AUC requires some positive and some
        # negative labels, so we should get one NULL value
        # for this config
        training_metric_groups = [{"metrics": ["accuracy", "roc_auc"]}]

        # Acquire fake data and objects to be used in the tests
        model_evaluator = ModelEvaluator(
            {},
            training_metric_groups,
            db_engine_with_results_schema,
        )
        fake_matrix_store = MockMatrixStore(
            matrix_type="train",
            matrix_uuid=str(labels),
            label_count=num_entities,
            db_engine=db_engine_with_results_schema,
            init_labels=pd.DataFrame(
                {
                    "label_value": labels,
                    "entity_id": list(range(num_entities)),
                    "as_of_date": [TRAIN_END_TIME] * num_entities,
                }
            ).set_index(["entity_id", "as_of_date"]).label_value,
            init_as_of_dates=[TRAIN_END_TIME],
        )

        model_evaluator.evaluate(
            trained_model.predict_proba(
                labels)[:, 1], fake_matrix_store, model_id
        )

        for metric, best, worst, stochastic in db_engine_with_results_schema.execute(
            f"""select metric, best_value, worst_value, stochastic_value
            from train_results.evaluations
            where model_id = %s and
            evaluation_start_time = %s
            order by 1""",
            (
                model_id,
                fake_matrix_store.as_of_dates[0]
            ),
        ):
            if metric == "accuracy":
                assert best is not None
                assert worst is not None
                assert stochastic is not None
            else:
                assert best is None
                assert worst is None
                assert stochastic is None


def test_subset_labels_and_predictions(db_engine_with_results_schema):
    num_entities = 5
    labels = [0, 1, 0, 1, 0]
    predictions_proba = np.array([0.6, 0.4, 0.55, 0.70, 0.3])

    fake_matrix_store = MockMatrixStore(
        matrix_type="test",
        matrix_uuid="abcde",
        label_count=num_entities,
        db_engine=db_engine_with_results_schema,
        init_labels=pd.DataFrame(
            {
                "label_value": labels,
                "entity_id": list(range(num_entities)),
                "as_of_date": [TRAIN_END_TIME] * num_entities,
            }
        ).set_index(["entity_id", "as_of_date"]).label_value,
        init_as_of_dates=[TRAIN_END_TIME],
    )

    for subset in SUBSETS:
        if subset["name"] == "evens":
            expected_result = 3
        elif subset["name"] == "odds":
            expected_result = 2
        elif subset["name"] == "empty":
            expected_result = 0

        populate_subset_data(db_engine_with_results_schema,
                             subset, list(range(num_entities)))
        subset_labels, subset_predictions, subset_protected_df = subset_labels_and_predictions(
            subset_df=query_subset_table(
                db_engine_with_results_schema,
                fake_matrix_store.as_of_dates,
                get_subset_table_name(subset),
            ),
            predictions_proba=predictions_proba,
            labels=fake_matrix_store.labels,
            protected_df=pd.DataFrame(),
        )

        assert len(subset_labels) == expected_result
        assert len(subset_predictions) == expected_result


def test_evaluating_early_warning(db_engine_with_results_schema):
    num_entities = 10
    labels = [0, 1, 0, 1, 0, 1, 0, 1, 0, 1]

    # Set up testing configuration parameters
    testing_metric_groups = [
        {
            "metrics": [
                "precision@",
                "recall@",
                "true positives@",
                "true negatives@",
                "false positives@",
                "false negatives@",
            ],
            "thresholds": {"percentiles": [5.0, 10.0], "top_n": [5, 10]},
        },
        {
            "metrics": [
                "f1",
                "mediocre",
                "accuracy",
                "roc_auc",
                "average precision score",
            ]
        },
        {"metrics": ["fbeta@"],
            "parameters": [{"beta": 0.75}, {"beta": 1.25}]},
    ]

    training_metric_groups = [{"metrics": ["accuracy", "roc_auc"]}]

    custom_metrics = {"mediocre": always_half}

    # Acquire fake data and objects to be used in the tests
    model_evaluator = ModelEvaluator(
        testing_metric_groups,
        training_metric_groups,
        db_engine_with_results_schema,
        custom_metrics=custom_metrics,
    )

    fake_test_matrix_store = MockMatrixStore(
        matrix_type="test",
        matrix_uuid="efgh",
        label_count=num_entities,
        db_engine=db_engine_with_results_schema,
        init_labels=pd.DataFrame(
            {
                "label_value": labels,
                "entity_id": list(range(num_entities)),
                "as_of_date": [TRAIN_END_TIME] * num_entities,
            }
        ).set_index(["entity_id", "as_of_date"]).label_value,
        init_as_of_dates=[TRAIN_END_TIME],
    )
    fake_train_matrix_store = MockMatrixStore(
        matrix_type="train",
        matrix_uuid="1234",
        label_count=num_entities,
        db_engine=db_engine_with_results_schema,
        init_labels=pd.DataFrame(
            {
                "label_value": labels,
                "entity_id": list(range(num_entities)),
                "as_of_date": [TRAIN_END_TIME] * num_entities,
            }
        ).set_index(["entity_id", "as_of_date"]).label_value,
        init_as_of_dates=[TRAIN_END_TIME],
    )

    trained_model, model_id = fake_trained_model(
        db_engine_with_results_schema,
        train_end_time=TRAIN_END_TIME,
    )

    # ensure that the matrix uuid is present
    matrix_uuids = [
        row[0]
        for row in db_engine_with_results_schema.execute("select matrix_uuid from test_results.evaluations")
    ]
    assert all(matrix_uuid == "efgh" for matrix_uuid in matrix_uuids)

    # Evaluate the training metrics and test
    model_evaluator.evaluate(
        trained_model.predict_proba(
            labels)[:, 1], fake_train_matrix_store, model_id
    )
    records = [
        row[0]
        for row in db_engine_with_results_schema.execute(
            """select distinct(metric || parameter)
            from train_results.evaluations
            where model_id = %s and
            evaluation_start_time = %s
            order by 1""",
            (model_id, fake_train_matrix_store.as_of_dates[0]),
        )
    ]
    assert records == ["accuracy", "roc_auc"]

    # Run tests for overall and subset evaluations
    for subset in SUBSETS:
        if subset is None:
            where_hash = ""
        else:
            populate_subset_data(db_engine_with_results_schema,
                                 subset, list(range(num_entities)))
            SubsetFactory(subset_hash=filename_friendly_hash(subset))
            session.commit()
            where_hash = f"and subset_hash = '{filename_friendly_hash(subset)}'"

        # Evaluate the testing metrics and test for all of them.
        model_evaluator.evaluate(
            trained_model.predict_proba(labels)[:, 1],
            fake_test_matrix_store,
            model_id,
            subset=subset,

        )

        records = [
            row[0]
            for row in db_engine_with_results_schema.execute(
                f"""\
                select distinct(metric || parameter)
                from test_results.evaluations
                where model_id = %s and
                evaluation_start_time = %s
                {where_hash}
                order by 1
                """,
                (
                    model_id,
                    fake_test_matrix_store.as_of_dates[0]
                ),
            )
        ]
        assert records == [
            "accuracy",
            "average precision score",
            "f1",
            "false negatives@10.0_pct",
            "false negatives@10_abs",
            "false negatives@5.0_pct",
            "false negatives@5_abs",
            "false positives@10.0_pct",
            "false positives@10_abs",
            "false positives@5.0_pct",
            "false positives@5_abs",
            "fbeta@0.75_beta",
            "fbeta@1.25_beta",
            "mediocre",
            "precision@10.0_pct",
            "precision@10_abs",
            "precision@5.0_pct",
            "precision@5_abs",
            "recall@10.0_pct",
            "recall@10_abs",
            "recall@5.0_pct",
            "recall@5_abs",
            "roc_auc",
            "true negatives@10.0_pct",
            "true negatives@10_abs",
            "true negatives@5.0_pct",
            "true negatives@5_abs",
            "true positives@10.0_pct",
            "true positives@10_abs",
            "true positives@5.0_pct",
            "true positives@5_abs",
        ]

        # Evaluate the training metrics and test
        model_evaluator.evaluate(
            trained_model.predict_proba(labels)[:, 1],
            fake_train_matrix_store,
            model_id,
            subset=subset,
        )

        records = [
            row[0]
            for row in db_engine_with_results_schema.execute(
                f"""select distinct(metric || parameter)
                from train_results.evaluations
                where model_id = %s and
                evaluation_start_time = %s
                {where_hash}
                order by 1""",
                (
                    model_id,
                    fake_train_matrix_store.as_of_dates[0]
                ),
            )
        ]
        assert records == ["accuracy", "roc_auc"]

    # ensure that the matrix uuid is present
    matrix_uuids = [
        row[0]
        for row in db_engine_with_results_schema.execute("select matrix_uuid from train_results.evaluations")
    ]
    assert all(matrix_uuid == "1234" for matrix_uuid in matrix_uuids)


def test_model_scoring_inspections(db_engine_with_results_schema):
    testing_metric_groups = [
        {
            "metrics": ["precision@", "recall@", "fpr@"],
            "thresholds": {"percentiles": [50.0], "top_n": [3]},
        },
        {
            # ensure we test a non-thresholded metric as well
            "metrics": ["accuracy"]
        },
    ]
    training_metric_groups = [
        {"metrics": ["accuracy"], "thresholds": {"percentiles": [50.0]}}
    ]

    model_evaluator = ModelEvaluator(
        testing_metric_groups,
        training_metric_groups,
        db_engine_with_results_schema,
    )

    testing_labels = np.array([1, 0, np.nan, 1, 0])
    testing_prediction_probas = np.array([0.56, 0.4, 0.55, 0.5, 0.3])

    training_labels = np.array(
        [0, 0, 1, 1, 1, 0, 1, 1]
    )
    training_prediction_probas = np.array(
        [0.6, 0.4, 0.55, 0.70, 0.3, 0.2, 0.8, 0.6]
    )

    fake_train_matrix_store = MockMatrixStore(
        "train", "efgh", 5, db_engine_with_results_schema, training_labels
    )
    fake_test_matrix_store = MockMatrixStore(
        "test", "1234", 5, db_engine_with_results_schema, testing_labels
    )

    trained_model, model_id = fake_trained_model(
        db_engine_with_results_schema,
        train_end_time=TRAIN_END_TIME,
    )

    # Evaluate testing matrix and test the results
    model_evaluator.evaluate(
        testing_prediction_probas, fake_test_matrix_store, model_id
    )
    for record in db_engine_with_results_schema.execute(
        """select * from test_results.evaluations
        where model_id = %s and evaluation_start_time = %s
        order by 1""",
        (model_id, fake_test_matrix_store.as_of_dates[0]),
    ):
        assert record["num_labeled_examples"] == 4
        assert record["num_positive_labels"] == 2
        if record["parameter"] == "":
            assert record["num_labeled_above_threshold"] == 4
        elif "pct" in record["parameter"]:
            assert record["num_labeled_above_threshold"] == 1
        else:
            assert record["num_labeled_above_threshold"] == 2

    # Evaluate the training matrix and test the results
    model_evaluator.evaluate(
        training_prediction_probas, fake_train_matrix_store, model_id
    )
    for record in db_engine_with_results_schema.execute(
        """select * from train_results.evaluations
        where model_id = %s and evaluation_start_time = %s
        order by 1""",
        (model_id, fake_train_matrix_store.as_of_dates[0]),
    ):
        assert record["num_labeled_examples"] == 8
        assert record["num_positive_labels"] == 5
        assert record["worst_value"] == 0.625
        assert record["best_value"] == 0.625
        assert record["stochastic_value"] == 0.625
        # best/worst are same, should shortcut trials
        assert record["num_sort_trials"] == 0
        assert record["standard_deviation"] == 0


def test_evaluation_with_sort_ties(db_engine_with_results_schema):
    model_evaluator = ModelEvaluator(
        testing_metric_groups=[
            {
                "metrics": ["precision@"],
                "thresholds": {"top_n": [3]},
            },
        ],
        training_metric_groups=[],
        db_engine=db_engine_with_results_schema,
    )
    testing_labels = np.array([1, 0, 1, 0, 0])
    testing_prediction_probas = np.array([0.56, 0.55, 0.5, 0.5, 0.3])

    fake_test_matrix_store = MockMatrixStore(
        "test", "1234", 5, db_engine_with_results_schema, testing_labels
    )

    trained_model, model_id = fake_trained_model(
        db_engine_with_results_schema,
        train_end_time=TRAIN_END_TIME,
    )
    model_evaluator.evaluate(
        testing_prediction_probas, fake_test_matrix_store, model_id
    )
    for record in db_engine_with_results_schema.execute(
        """select * from test_results.evaluations
        where model_id = %s and evaluation_start_time = %s
        order by 1""",
        (model_id, fake_test_matrix_store.as_of_dates[0]),
    ):
        assert record["num_labeled_examples"] == 5
        assert record["num_positive_labels"] == 2
        assert_almost_equal(float(record["worst_value"]), 0.33333, 5)
        assert_almost_equal(float(record["best_value"]), 0.66666, 5)
        assert record["num_sort_trials"] == SORT_TRIALS
        assert record["stochastic_value"] > record["worst_value"]
        assert record["stochastic_value"] < record["best_value"]
        assert record["standard_deviation"]


def test_ModelEvaluator_needs_evaluation_no_bias_audit(db_engine_with_results_schema):
    # TEST SETUP:

    # create two models: one that has zero evaluations,
    # one that has an evaluation for precision@100_abs
    # both overall and for each subset
    model_with_evaluations = ModelFactory()
    model_without_evaluations = ModelFactory()

    eval_time = datetime.datetime(2016, 1, 1)
    as_of_date_frequency = "3d"
    for subset_hash in [""] + [filename_friendly_hash(subset) for subset in SUBSETS]:
        EvaluationFactory(
            model_rel=model_with_evaluations,
            evaluation_start_time=eval_time,
            evaluation_end_time=eval_time,
            as_of_date_frequency=as_of_date_frequency,
            metric="precision@",
            parameter="100_abs",
            subset_hash=subset_hash,
        )
    session.commit()

    # make a test matrix to pass in
    metadata_overrides = {
        'as_of_date_frequency': as_of_date_frequency,
        'as_of_times': [eval_time],
    }
    test_matrix_store = MockMatrixStore(
        "test", "1234", 5, db_engine_with_results_schema, metadata_overrides=metadata_overrides
    )
    train_matrix_store = MockMatrixStore(
        "train", "2345", 5, db_engine_with_results_schema, metadata_overrides=metadata_overrides
    )

    # the evaluated model has test evaluations for precision, but not recall,
    # so this needs evaluations
    for subset in SUBSETS:
        if not subset:
            subset_hash = ""
        else:
            subset_hash = filename_friendly_hash(subset)

        assert ModelEvaluator(
            testing_metric_groups=[{
                "metrics": ["precision@", "recall@"],
                "thresholds": {"top_n": [100]},
            }],
            training_metric_groups=[],
            db_engine=db_engine_with_results_schema,
        ).needs_evaluations(
            matrix_store=test_matrix_store,
            model_id=model_with_evaluations.model_id,
            subset_hash=subset_hash,
        )

    # the evaluated model has test evaluations for precision,
    # so this should not need evaluations
    for subset in SUBSETS:
        if not subset:
            subset_hash = ""
        else:
            subset_hash = filename_friendly_hash(subset)

        assert not ModelEvaluator(
            testing_metric_groups=[{
                "metrics": ["precision@"],
                "thresholds": {"top_n": [100]},
            }],
            training_metric_groups=[],
            db_engine=db_engine_with_results_schema,
        ).needs_evaluations(
            matrix_store=test_matrix_store,
            model_id=model_with_evaluations.model_id,
            subset_hash=subset_hash,
        )

    # the non-evaluated model has no evaluations,
    # so this should need evaluations
    for subset in SUBSETS:
        if not subset:
            subset_hash = ""
        else:
            subset_hash = filename_friendly_hash(subset)

        assert ModelEvaluator(
            testing_metric_groups=[{
                "metrics": ["precision@"],
                "thresholds": {"top_n": [100]},
            }],
            training_metric_groups=[],
            db_engine=db_engine_with_results_schema,
        ).needs_evaluations(
            matrix_store=test_matrix_store,
            model_id=model_without_evaluations.model_id,
            subset_hash=subset_hash,
        )

    # the evaluated model has no *train* evaluations,
    # so the train matrix should need evaluations
    for subset in SUBSETS:
        if not subset:
            subset_hash = ""
        else:
            subset_hash = filename_friendly_hash(subset)

        assert ModelEvaluator(
            testing_metric_groups=[{
                "metrics": ["precision@"],
                "thresholds": {"top_n": [100]},
            }],
            training_metric_groups=[{
                "metrics": ["precision@"],
                "thresholds": {"top_n": [100]},
            }],
            db_engine=db_engine_with_results_schema,
        ).needs_evaluations(
            matrix_store=train_matrix_store,
            model_id=model_with_evaluations.model_id,
            subset_hash=subset_hash,
        )
    session.close()
    session.remove()


def test_ModelEvaluator_needs_evaluation_with_bias_audit(db_engine_with_results_schema):
    # test that if a bias audit config is passed, and there are no matching bias audits
    # in the database, needs_evaluation returns true
    # this all assumes that evaluations are populated. those tests are in the 'no_bias_audit' test
    model_evaluator = ModelEvaluator(
        testing_metric_groups=[
            {
                "metrics": ["precision@"],
                "thresholds": {"top_n": [3]},
            },
        ],
        training_metric_groups=[],
        bias_config={
            'thresholds': {'top_n': [2]}
        },
        db_engine=db_engine_with_results_schema,
    )
    model_with_evaluations = ModelFactory()

    eval_time = datetime.datetime(2016, 1, 1)
    as_of_date_frequency = "3d"
    for subset_hash in [""]:
        EvaluationFactory(
            model_rel=model_with_evaluations,
            evaluation_start_time=eval_time,
            evaluation_end_time=eval_time,
            as_of_date_frequency=as_of_date_frequency,
            metric="precision@",
            parameter="3_abs",
            subset_hash=subset_hash,
        )
    session.commit()

    # make a test matrix to pass in
    metadata_overrides = {
        'as_of_date_frequency': as_of_date_frequency,
        'as_of_times': [eval_time],
    }
    test_matrix_store = MockMatrixStore(
        "test", "1234", 5, db_engine_with_results_schema, metadata_overrides=metadata_overrides
    )
    assert model_evaluator.needs_evaluations(
        matrix_store=test_matrix_store,
        model_id=model_with_evaluations.model_id,
        subset_hash="",
    )


def test_evaluation_with_protected_df(db_engine_with_results_schema):
    # Test that if a protected_df is passed (along with bias config, the only real needed one
    # being threshold info), an Aequitas report is written to the database.
    model_evaluator = ModelEvaluator(
        testing_metric_groups=[
            {
                "metrics": ["precision@"],
                "thresholds": {"top_n": [3]},
            },
        ],
        training_metric_groups=[],
        bias_config={
            'thresholds': {'top_n': [2]}
        },
        db_engine=db_engine_with_results_schema,
    )
    testing_labels = np.array([1, 0])
    testing_prediction_probas = np.array([0.56, 0.55])

    fake_test_matrix_store = MockMatrixStore(
        "test", "1234", 5, db_engine_with_results_schema, testing_labels
    )

    trained_model, model_id = fake_trained_model(
        db_engine_with_results_schema,
        train_end_time=TRAIN_END_TIME,
    )

    protected_df = pd.DataFrame({
        "entity_id": fake_test_matrix_store.design_matrix.index.levels[0].tolist(),
        "protectedattribute1": "value1"
    })

    model_evaluator.evaluate(
        testing_prediction_probas, fake_test_matrix_store, model_id, protected_df
    )
    for record in db_engine_with_results_schema.execute(
        """select * from test_results.aequitas
        where model_id = %s and evaluation_start_time = %s
        order by 1""",
        (model_id, fake_test_matrix_store.as_of_dates[0]),
    ):
        assert record['model_id'] == model_id
        assert record['parameter'] == '2_abs'
        assert record['attribute_name'] == 'protectedattribute1'
        assert record['attribute_value'] == 'value1'


def test_evaluation_sorting_with_protected_df(db_engine_with_results_schema):
    # Test that if a protected_df is passed (along with bias config, the only real needed one
    # being threshold info), an Aequitas report is written to the database.
    model_evaluator = ModelEvaluator(
        testing_metric_groups=[
            {
                "metrics": ["precision@"],
                "thresholds": {"top_n": [3]},
            },
        ],
        training_metric_groups=[],
        bias_config={
            'thresholds': {'top_n': [2]}
        },
        db_engine=db_engine_with_results_schema,
    )
    testing_labels = np.array([1, 1, 1, 0, 1])
    testing_prediction_probas = np.array([0.56, 0.55, 0.92, 0.85, 0.24])

    fake_test_matrix_store = MockMatrixStore(
        "test", "1234", 5, db_engine_with_results_schema,
        metadata_overrides={'as_of_times': [TRAIN_END_TIME]},
        matrix=pd.DataFrame.from_dict(
                {
                    "entity_id": [1, 2, 3, 4, 5],
                    "as_of_date": [pd.Timestamp(2016, 1, 1)]*5,
                    "feature_one": [3, 4, 3, 4, 3],
                    "feature_two": [5, 6, 5, 6, 5],
                    "label": testing_labels,
                }
            ).set_index(MatrixStore.indices),
        init_labels=pd.DataFrame(
            {
                "label_value": testing_labels,
                "entity_id": [1, 2, 3, 4, 5],
                "as_of_date": [pd.Timestamp(2016, 1, 1)]*5,
            }
        ).set_index(["entity_id", "as_of_date"]).label_value,
        init_as_of_dates=[TRAIN_END_TIME]
    )

    trained_model, model_id = fake_trained_model(
        db_engine_with_results_schema,
        train_end_time=TRAIN_END_TIME,
    )

    protected_df = pd.DataFrame({
        # "entity_id": fake_test_matrix_store.design_matrix.index.levels[0].tolist(),
        # "as_of_date": fake_test_matrix_store.design_matrix.index.levels[1].tolist(),
        "protectedattribute1": ["low", "low", "low", "high", "high"]
    }, index=fake_test_matrix_store.design_matrix.index)
    # should be low has 3 records, all 1's; high has 2 records, one 1

    expected = {
        "low": {"group_size": 3, "group_label_neg": 0, "group_label_pos": 3},
        "high": {"group_size": 2, "group_label_neg": 1, "group_label_pos": 1}
    }

    model_evaluator.evaluate(
        testing_prediction_probas, fake_test_matrix_store, model_id, protected_df
    )

    for record in db_engine_with_results_schema.execute(
        """select * from test_results.aequitas
        where model_id = %s and evaluation_start_time = %s
        order by 1""",
        (model_id, fake_test_matrix_store.as_of_dates[0]),
    ):
        assert record['model_id'] == model_id
        assert record['parameter'] == '2_abs'
        assert record['attribute_name'] == 'protectedattribute1'
        for col, value in expected[record['attribute_value']].items():
            assert record[col] == value



def test_generate_binary_at_x():
    input_array = np.array(
        [0.9, 0.8, 0.7, 0.7, 0.7, 0.7, 0.7, 0.7, 0.7, 0.6])

    # bug can arise when the same value spans both sides of threshold
    assert_array_equal(
        generate_binary_at_x(input_array, 50, "percentile"),
        np.array([1, 1, 1, 1, 1, 0, 0, 0, 0, 0])
    )

    assert_array_equal(
        generate_binary_at_x(input_array, 2),
        np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0])
    )

    assert_array_equal(
        generate_binary_at_x(np.array([]), 2),
        np.array([])
    )



"""add_matrix_uuid_to_eval

Revision ID: 0bca1ba9706e
Revises: 38f37d013686
Create Date: 2019-02-05 13:19:50.172109

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '0bca1ba9706e'
down_revision = '38f37d013686'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('evaluations', sa.Column('matrix_uuid', sa.Text(), nullable=True), schema='test_results')
    op.create_foreign_key(None, 'evaluations', 'matrices', ['matrix_uuid'], ['matrix_uuid'], source_schema='test_results', referent_schema='model_metadata')
    op.add_column('evaluations', sa.Column('matrix_uuid', sa.Text(), nullable=True), schema='train_results')
    op.create_foreign_key(None, 'evaluations', 'matrices', ['matrix_uuid'], ['matrix_uuid'], source_schema='train_results', referent_schema='model_metadata')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'evaluations', schema='train_results', type_='foreignkey')
    op.drop_column('evaluations', 'matrix_uuid', schema='train_results')
    op.drop_constraint(None, 'evaluations', schema='test_results', type_='foreignkey')
    op.drop_column('evaluations', 'matrix_uuid', schema='test_results')
    # ### end Alembic commands ###



import pytest

from triage.component.catwalk.feature_importances import (
    get_feature_importances,
)

from sklearn import datasets
from sklearn.svm import SVC
from sklearn.dummy import DummyClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression

from sklearn.model_selection import train_test_split


@pytest.fixture
def trained_models():
    dataset = datasets.load_breast_cancer()
    X = dataset.data
    y = dataset.target

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=12345
    )

    rf = RandomForestClassifier(n_estimators=100)
    rf.fit(X_train, y_train)

    lr = LogisticRegression(solver='liblinear')
    lr.fit(X_train, y_train)

    svc_w_linear_kernel = SVC(kernel="linear", gamma='auto')
    svc_w_linear_kernel.fit(X_train, y_train)

    svc_wo_linear_kernel = SVC(gamma='auto')
    svc_wo_linear_kernel.fit(X_train, y_train)

    dummy = DummyClassifier(strategy='stratified')
    dummy.fit(X_train, y_train)

    return {
        "RF": rf,
        "LR": lr,
        "SVC_w_linear_kernel": svc_w_linear_kernel,
        "Dummy": dummy,
        "SVC_wo_linear_kernel": svc_wo_linear_kernel,
    }

def test_correct_feature_importances_for_lr(trained_models):
    feature_importances = get_feature_importances(trained_models["LR"])

    # It returns the intercept, too
    assert feature_importances.shape == (30,)


def test_correct_feature_importances_for_rf(trained_models):
    feature_importances = get_feature_importances(trained_models["RF"])
    assert feature_importances.shape == (30,)


def test_correct_feature_importances_for_svc_w_linear_kernel(trained_models):
    feature_importances = get_feature_importances(
        trained_models["SVC_w_linear_kernel"])
    assert feature_importances.shape == (30,)


def test_correct_feature_importances_for_svc_wo_linear_kernel(trained_models):
    feature_importances = get_feature_importances(
        trained_models["SVC_wo_linear_kernel"]
    )
    assert feature_importances is None


def test_correct_feature_importances_for_dummy(trained_models):
    feature_importances = get_feature_importances(trained_models["Dummy"])
    assert feature_importances is None



import datetime
import shutil
import sys
import tempfile
import random
from contextlib import contextmanager

import pandas as pd
import yaml
import numpy as np


def convert_string_column_to_date(column):
    return [datetime.datetime.strptime(date, "%Y-%m-%d").date() for date in column]


def create_schemas(engine, features_tables, labels, states):
    """ This function makes a features schema and populates it with the fake
    data from above.

    :param engine: a postgresql engine
    :type engine: sqlalchemy.engine
    """
    # create features schema and associated tables
    engine.execute("drop schema if exists features cascade; create schema features;")
    for table_number, table in enumerate(features_tables):
        create_features_table(table_number, table, engine)
    # create labels schema and table
    engine.execute("drop schema if exists labels cascade; create schema labels;")
    engine.execute(
        """
            create table labels.labels (
                entity_id int,
                as_of_date date,
                label_timespan interval,
                label_name char(30),
                label_type char(30),
                label int
            )
        """
    )
    for row in labels:
        engine.execute("insert into labels.labels values (%s, %s, %s, %s, %s, %s)", row)

    # create cohort table
    engine.execute("drop schema if exists staging cascade; create schema staging;")
    engine.execute(
        """
            create table cohort (
                entity_id int,
                as_of_date date,
                active bool
            )
        """
    )
    for row in states:
        engine.execute("insert into cohort values (%s, %s, %s)", row)


def create_features_table(table_number, table, engine):
    engine.execute(
        """
            create table features.features{} (
                entity_id int, as_of_date date, f{} int, f{} int
            )
        """.format(
            table_number, (table_number * 2) + 1, (table_number * 2) + 2
        )
    )
    for row in table:
        engine.execute(
            """
                insert into features.features{} values (%s, %s, %s, %s)
            """.format(
                table_number
            ),
            row,
        )


def create_entity_date_df(
    labels,
    states,
    as_of_dates,
    label_name,
    label_type,
    label_timespan,
):
    """ This function makes a pandas DataFrame that mimics the entity-date table
    for testing against.
    """
    0, "2016-02-01", "1 month", "booking", "binary", 0
    labels_table = pd.DataFrame(
        labels,
        columns=[
            "entity_id",
            "as_of_date",
            "label_timespan",
            "label_name",
            "label_type",
            "label",
        ],
    )
    states_table = pd.DataFrame(
        states, columns=["entity_id", "as_of_date", "active"]
    ).set_index(["entity_id", "as_of_date"])
    as_of_dates = [date.date() for date in as_of_dates]
    labels_table = labels_table[labels_table["label_name"] == label_name]
    labels_table = labels_table[labels_table["label_type"] == label_type]
    labels_table = labels_table[labels_table["label_timespan"] == label_timespan]
    labels_table = labels_table.join(other=states_table, on=("entity_id", "as_of_date"))
    labels_table = labels_table[labels_table["active"]]
    ids_dates = labels_table[["entity_id", "as_of_date"]]
    ids_dates = ids_dates.sort_values(["entity_id", "as_of_date"])
    ids_dates["as_of_date"] = [
        datetime.datetime.strptime(date, "%Y-%m-%d").date()
        for date in ids_dates["as_of_date"]
    ]
    ids_dates = ids_dates[ids_dates["as_of_date"].isin(as_of_dates)]
    print(ids_dates)

    return ids_dates.reset_index(drop=True)


def NamedTempFile():
    if sys.version_info >= (3, 0, 0):
        return tempfile.NamedTemporaryFile(mode="w+", newline="")
    else:
        return tempfile.NamedTemporaryFile()


@contextmanager
def TemporaryDirectory():
    name = tempfile.mkdtemp()
    try:
        yield name
    finally:
        shutil.rmtree(name)


def fake_labels(length):
    return np.array([random.choice([True, False]) for i in range(0, length)])


def assert_index(engine, table, column):
    """Assert that a table has an index on a given column

    Does not care which position the column is in the index
    Modified from https://www.gab.lc/articles/index_on_id_with_postgresql

    Args:
        engine (sqlalchemy.engine) a database engine
        table (string) the name of a table
        column (string) the name of a column
    """
    query = """
        SELECT 1
        FROM pg_class t
             JOIN pg_index ix ON t.oid = ix.indrelid
             JOIN pg_class i ON i.oid = ix.indexrelid
             JOIN pg_attribute a ON a.attrelid = t.oid
        WHERE
             a.attnum = ANY(ix.indkey) AND
             t.relkind = 'r' AND
             t.relname = '{table_name}' AND
             a.attname = '{column_name}'
    """.format(
        table_name=table, column_name=column
    )
    num_results = len([row for row in engine.execute(query)])
    assert num_results >= 1


def create_binary_outcome_events(db_engine, table_name, events_data):
    db_engine.execute(
        "create table events (entity_id int, outcome_date date, outcome bool)"
    )
    for event in events_data:
        db_engine.execute(
            "insert into {} values (%s, %s, %s::bool)".format(table_name), event
        )



import datetime
from unittest import TestCase

import pandas as pd
import testing.postgresql
from mock import Mock
from triage import create_engine
from contextlib import contextmanager

from triage.component.catwalk.utils import filename_friendly_hash
from triage.component.architect.feature_group_creator import FeatureGroup
from triage.component.architect.builders import MatrixBuilder
from triage.component.catwalk.db import ensure_db
from triage.component.catwalk.storage import ProjectStorage
from triage.component.results_schema.schema import Matrix

from .utils import (
    create_schemas,
    create_entity_date_df,
    convert_string_column_to_date,
    TemporaryDirectory,
)

from tests.utils import matrix_metadata_creator


# make some fake features data

states = [
    [0, "2016-01-01", False],
    [0, "2016-02-01", False],
    [0, "2016-03-01", False],
    [0, "2016-04-01", False],
    [0, "2016-05-01", False],
    [0, "2016-06-01", True],
    [1, "2016-01-01", False],
    [1, "2016-02-01", False],
    [1, "2016-03-01", False],
    [1, "2016-04-01", False],
    [1, "2016-05-01", False],
    [2, "2016-01-01", False],
    [2, "2016-02-01", True],
    [2, "2016-03-01", False],
    [2, "2016-04-01", True],
    [2, "2016-05-01", False],
    [3, "2016-01-01", False],
    [3, "2016-02-01", True],
    [3, "2016-03-01", False],
    [3, "2016-04-01", True],
    [3, "2016-05-01", False],
    [4, "2016-01-01", True],
    [4, "2016-02-01", True],
    [4, "2016-03-01", True],
    [4, "2016-04-01", True],
    [4, "2016-05-01", True],
    [5, "2016-01-01", False],
    [5, "2016-02-01", False],
    [5, "2016-03-01", False],
    [5, "2016-04-01", False],
    [5, "2016-05-01", False],
]

features0_pre = [
    [0, "2016-01-01", 2, 1],
    [1, "2016-01-01", 1, 2],
    [0, "2016-02-01", 2, 3],
    [1, "2016-02-01", 2, 4],
    [0, "2016-03-01", 3, 3],
    [1, "2016-03-01", 3, 4],
    [0, "2016-04-01", 4, 3],
    [1, "2016-05-01", 5, 4],
]

features1_pre = [
    [2, "2016-01-01", 1, 1],
    [3, "2016-01-01", 1, 2],
    [2, "2016-02-01", 2, 3],
    [3, "2016-02-01", 2, 2],
    [0, "2016-03-01", 3, 3],
    [1, "2016-03-01", 3, 4],
    [2, "2016-03-01", 3, 3],
    [3, "2016-03-01", 3, 4],
    [3, "2016-03-01", 3, 4],
    [0, "2016-03-01", 3, 3],
    [4, "2016-03-01", 1, 4],
    [5, "2016-03-01", 2, 4],
]

# collate will ensure every entity/date combination in the state
# table have an imputed value in the features table, so ensure
# this is true for our test (filling with 9's):
f0_dict = {(r[0], r[1]): r for r in features0_pre}
f1_dict = {(r[0], r[1]): r for r in features1_pre}

for rec in states:
    ent_dt = (rec[0], rec[1])
    f0_dict[ent_dt] = f0_dict.get(ent_dt, list(ent_dt + (9, 9)))
    f1_dict[ent_dt] = f1_dict.get(ent_dt, list(ent_dt + (9, 9)))

features0 = sorted(f0_dict.values(), key=lambda x: (x[1], x[0]))
features1 = sorted(f1_dict.values(), key=lambda x: (x[1], x[0]))

features_tables = [features0, features1]

# make some fake labels data

labels = [
    [0, "2016-02-01", "1 month", "booking", "binary", 0],
    [0, "2016-03-01", "1 month", "booking", "binary", 0],
    [0, "2016-04-01", "1 month", "booking", "binary", 0],
    [0, "2016-05-01", "1 month", "booking", "binary", 1],
    [0, "2016-01-01", "1 month", "ems", "binary", 0],
    [0, "2016-02-01", "1 month", "ems", "binary", 0],
    [0, "2016-03-01", "1 month", "ems", "binary", 0],
    [0, "2016-04-01", "1 month", "ems", "binary", 0],
    [0, "2016-05-01", "1 month", "ems", "binary", 0],
    [1, "2016-01-01", "1 month", "booking", "binary", 0],
    [1, "2016-02-01", "1 month", "booking", "binary", 0],
    [1, "2016-03-01", "1 month", "booking", "binary", 0],
    [1, "2016-04-01", "1 month", "booking", "binary", 0],
    [1, "2016-05-01", "1 month", "booking", "binary", 1],
    [1, "2016-01-01", "1 month", "ems", "binary", 0],
    [1, "2016-02-01", "1 month", "ems", "binary", 0],
    [1, "2016-03-01", "1 month", "ems", "binary", 0],
    [1, "2016-04-01", "1 month", "ems", "binary", 0],
    [1, "2016-05-01", "1 month", "ems", "binary", 0],
    [2, "2016-01-01", "1 month", "booking", "binary", 0],
    [2, "2016-02-01", "1 month", "booking", "binary", 0],
    [2, "2016-03-01", "1 month", "booking", "binary", 1],
    [2, "2016-04-01", "1 month", "booking", "binary", 0],
    [2, "2016-05-01", "1 month", "booking", "binary", 1],
    [2, "2016-01-01", "1 month", "ems", "binary", 0],
    [2, "2016-02-01", "1 month", "ems", "binary", 0],
    [2, "2016-03-01", "1 month", "ems", "binary", 0],
    [2, "2016-04-01", "1 month", "ems", "binary", 0],
    [2, "2016-05-01", "1 month", "ems", "binary", 1],
    [3, "2016-01-01", "1 month", "booking", "binary", 0],
    [3, "2016-02-01", "1 month", "booking", "binary", 0],
    [3, "2016-03-01", "1 month", "booking", "binary", 1],
    [3, "2016-04-01", "1 month", "booking", "binary", 0],
    [3, "2016-05-01", "1 month", "booking", "binary", 1],
    [3, "2016-01-01", "1 month", "ems", "binary", 0],
    [3, "2016-02-01", "1 month", "ems", "binary", 0],
    [3, "2016-03-01", "1 month", "ems", "binary", 0],
    [3, "2016-04-01", "1 month", "ems", "binary", 1],
    [3, "2016-05-01", "1 month", "ems", "binary", 0],
    [4, "2016-01-01", "1 month", "booking", "binary", 1],
    [4, "2016-02-01", "1 month", "booking", "binary", 0],
    [4, "2016-03-01", "1 month", "booking", "binary", 0],
    [4, "2016-04-01", "1 month", "booking", "binary", 0],
    [4, "2016-05-01", "1 month", "booking", "binary", 0],
    [4, "2016-01-01", "1 month", "ems", "binary", 0],
    [4, "2016-02-01", "1 month", "ems", "binary", 1],
    [4, "2016-03-01", "1 month", "ems", "binary", 0],
    [4, "2016-04-01", "1 month", "ems", "binary", 1],
    [4, "2016-05-01", "1 month", "ems", "binary", 1],
    [5, "2016-01-01", "1 month", "booking", "binary", 1],
    [5, "2016-02-01", "1 month", "booking", "binary", 0],
    [5, "2016-03-01", "1 month", "booking", "binary", 0],
    [5, "2016-04-01", "1 month", "booking", "binary", 0],
    [5, "2016-05-01", "1 month", "booking", "binary", 0],
    [5, "2016-01-01", "1 month", "ems", "binary", 0],
    [5, "2016-02-01", "1 month", "ems", "binary", 1],
    [5, "2016-03-01", "1 month", "ems", "binary", 0],
    [5, "2016-04-01", "1 month", "ems", "binary", 0],
    [5, "2016-05-01", "1 month", "ems", "binary", 0],
    [0, "2016-02-01", "3 month", "booking", "binary", 0],
    [0, "2016-03-01", "3 month", "booking", "binary", 0],
    [0, "2016-04-01", "3 month", "booking", "binary", 0],
    [0, "2016-05-01", "3 month", "booking", "binary", 1],
    [0, "2016-01-01", "3 month", "ems", "binary", 0],
    [0, "2016-02-01", "3 month", "ems", "binary", 0],
    [0, "2016-03-01", "3 month", "ems", "binary", 0],
    [0, "2016-04-01", "3 month", "ems", "binary", 0],
    [0, "2016-05-01", "3 month", "ems", "binary", 0],
    [1, "2016-01-01", "3 month", "booking", "binary", 0],
    [1, "2016-02-01", "3 month", "booking", "binary", 0],
    [1, "2016-03-01", "3 month", "booking", "binary", 0],
    [1, "2016-04-01", "3 month", "booking", "binary", 0],
    [1, "2016-05-01", "3 month", "booking", "binary", 1],
    [1, "2016-01-01", "3 month", "ems", "binary", 0],
    [1, "2016-02-01", "3 month", "ems", "binary", 0],
    [1, "2016-03-01", "3 month", "ems", "binary", 0],
    [1, "2016-04-01", "3 month", "ems", "binary", 0],
    [1, "2016-05-01", "3 month", "ems", "binary", 0],
    [2, "2016-01-01", "3 month", "booking", "binary", 0],
    [2, "2016-02-01", "3 month", "booking", "binary", 0],
    [2, "2016-03-01", "3 month", "booking", "binary", 1],
    [2, "2016-04-01", "3 month", "booking", "binary", 0],
    [2, "2016-05-01", "3 month", "booking", "binary", 1],
    [2, "2016-01-01", "3 month", "ems", "binary", 0],
    [2, "2016-02-01", "3 month", "ems", "binary", 0],
    [2, "2016-03-01", "3 month", "ems", "binary", 0],
    [2, "2016-04-01", "3 month", "ems", "binary", 0],
    [2, "2016-05-01", "3 month", "ems", "binary", 1],
    [3, "2016-01-01", "3 month", "booking", "binary", 0],
    [3, "2016-02-01", "3 month", "booking", "binary", 0],
    [3, "2016-03-01", "3 month", "booking", "binary", 1],
    [3, "2016-04-01", "3 month", "booking", "binary", 0],
    [3, "2016-05-01", "3 month", "booking", "binary", 1],
    [3, "2016-01-01", "3 month", "ems", "binary", 0],
    [3, "2016-02-01", "3 month", "ems", "binary", 0],
    [3, "2016-03-01", "3 month", "ems", "binary", 0],
    [3, "2016-04-01", "3 month", "ems", "binary", 1],
    [3, "2016-05-01", "3 month", "ems", "binary", 0],
    [3, "2016-05-01", "3 month", "ems", "binary", 0],
    [4, "2016-01-01", "3 month", "booking", "binary", 0],
    [4, "2016-02-01", "3 month", "booking", "binary", 0],
    [4, "2016-03-01", "3 month", "booking", "binary", 1],
    [4, "2016-04-01", "3 month", "booking", "binary", 0],
    [4, "2016-05-01", "3 month", "booking", "binary", 1],
    [4, "2016-01-01", "3 month", "ems", "binary", 0],
    [4, "2016-02-01", "3 month", "ems", "binary", 0],
    [4, "2016-03-01", "3 month", "ems", "binary", 0],
    [4, "2016-04-01", "3 month", "ems", "binary", 0],
    [4, "2016-05-01", "3 month", "ems", "binary", 1],
    [5, "2016-01-01", "3 month", "booking", "binary", 0],
    [5, "2016-02-01", "3 month", "booking", "binary", 0],
    [5, "2016-03-01", "3 month", "booking", "binary", 1],
    [5, "2016-04-01", "3 month", "booking", "binary", 0],
    [5, "2016-05-01", "3 month", "booking", "binary", 1],
    [5, "2016-01-01", "3 month", "ems", "binary", 0],
    [5, "2016-02-01", "3 month", "ems", "binary", 0],
    [5, "2016-03-01", "3 month", "ems", "binary", 0],
    [5, "2016-04-01", "3 month", "ems", "binary", 1],
    [5, "2016-05-01", "3 month", "ems", "binary", 0],
]

label_name = "booking"
label_type = "binary"

db_config = {
    "features_schema_name": "features",
    "labels_schema_name": "labels",
    "labels_table_name": "labels",
    "cohort_table_name": "cohort",
}

experiment_hash = None


@contextmanager
def get_matrix_storage_engine():
    with TemporaryDirectory() as temp_dir:
        yield ProjectStorage(temp_dir).matrix_storage_engine()


def test_query_to_df():
    """ Test the write_to_csv function by checking whether the csv contains the
    correct number of lines.
    """
    with testing.postgresql.Postgresql() as postgresql:
        # create an engine and generate a table with fake feature data
        engine = create_engine(postgresql.url())
        create_schemas(
            engine=engine, features_tables=features_tables, labels=labels, states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                experiment_hash=experiment_hash,
                engine=engine,
            )

            # for each table, check that corresponding csv has the correct # of rows
            for table in features_tables:
                df = builder.query_to_df(
                    """
                        select *
                        from features.features{}
                    """.format(
                        features_tables.index(table)
                    )
                )
                assert len(df) == len(table)


def test_make_entity_date_table():
    """ Test that the make_entity_date_table function contains the correct
    values.
    """
    dates = [
        datetime.datetime(2016, 1, 1, 0, 0),
        datetime.datetime(2016, 2, 1, 0, 0),
        datetime.datetime(2016, 3, 1, 0, 0),
    ]

    # make a dataframe of entity ids and dates to test against
    ids_dates = create_entity_date_df(
        labels=labels,
        states=states,
        as_of_dates=dates,
        label_name="booking",
        label_type="binary",
        label_timespan="1 month",
    )

    with testing.postgresql.Postgresql() as postgresql:
        # create an engine and generate a table with fake feature data
        engine = create_engine(postgresql.url())
        create_schemas(
            engine=engine, features_tables=features_tables, labels=labels, states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                experiment_hash=experiment_hash,
                engine=engine,
            )
            engine.execute("CREATE TABLE features.tmp_entity_date (a int, b date);")
            # call the function to test the creation of the table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type="binary",
                label_name="booking",
                state="active",
                matrix_uuid="my_uuid",
                matrix_type="train",
                label_timespan="1 month",
            )

            # read in the table
            result = pd.read_sql(
                "select * from features.{} order by entity_id, as_of_date".format(
                    entity_date_table_name
                ),
                engine,
            )
            # compare the table to the test dataframe
            test = result == ids_dates
            assert test.all().all()


def test_make_entity_date_table_include_missing_labels():
    """ Test that the make_entity_date_table function contains the correct
    values.
    """
    dates = [
        datetime.datetime(2016, 1, 1, 0, 0),
        datetime.datetime(2016, 2, 1, 0, 0),
        datetime.datetime(2016, 3, 1, 0, 0),
        datetime.datetime(2016, 6, 1, 0, 0),
    ]

    # same as the other make_entity_date_label test except there is an extra date, 2016-06-01
    # entity 0 is included in this date via the states table, but has no label

    # make a dataframe of entity ids and dates to test against
    ids_dates = create_entity_date_df(
        labels=labels,
        states=states,
        as_of_dates=dates,
        label_name="booking",
        label_type="binary",
        label_timespan="1 month",
    )
    # this line adds the new entity-date combo as an expected one
    ids_dates = ids_dates.append(
        {"entity_id": 0, "as_of_date": datetime.date(2016, 6, 1)}, ignore_index=True
    )

    with testing.postgresql.Postgresql() as postgresql:
        # create an engine and generate a table with fake feature data
        engine = create_engine(postgresql.url())
        create_schemas(
            engine=engine, features_tables=features_tables, labels=labels, states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                experiment_hash=experiment_hash,
                include_missing_labels_in_train_as=False,
                engine=engine,
            )
            engine.execute("CREATE TABLE features.tmp_entity_date (a int, b date);")
            # call the function to test the creation of the table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type="binary",
                label_name="booking",
                state="active",
                matrix_uuid="my_uuid",
                matrix_type="train",
                label_timespan="1 month",
            )

            # read in the table
            result = pd.read_sql(
                "select * from features.{} order by entity_id, as_of_date".format(
                    entity_date_table_name
                ),
                engine,
            )

            # compare the table to the test dataframe
            assert sorted(result.values.tolist()) == sorted(ids_dates.values.tolist())


def test_load_features_data():
    dates = [datetime.datetime(2016, 1, 1, 0, 0), datetime.datetime(2016, 2, 1, 0, 0)]

    # make dataframe for entity ids and dates
    ids_dates = create_entity_date_df(
        labels=labels,
        states=states,
        as_of_dates=dates,
        label_name="booking",
        label_type="binary",
        label_timespan="1 month",
    )

    features = [["f1", "f2"], ["f3", "f4"]]
    # make dataframes of features to test against
    features_dfs = []
    for i, table in enumerate(features_tables):
        cols = ["entity_id", "as_of_date"] + features[i]
        temp_df = pd.DataFrame(table, columns=cols)
        temp_df["as_of_date"] = convert_string_column_to_date(temp_df["as_of_date"])
        features_dfs.append(
            ids_dates.merge(
                right=temp_df, how="left", on=["entity_id", "as_of_date"]
            ).set_index(["entity_id", "as_of_date"])
        )

    # create an engine and generate a table with fake feature data
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_schemas(
            engine=engine, features_tables=features_tables, labels=labels, states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                experiment_hash=experiment_hash,
                engine=engine,
            )

            # make the entity-date table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type="binary",
                label_name="booking",
                state="active",
                matrix_type="train",
                matrix_uuid="my_uuid",
                label_timespan="1 month",
            )

            feature_dictionary = dict(
                ("features{}".format(i), feature_list)
                for i, feature_list in enumerate(features)
            )

            returned_features_dfs = builder.load_features_data(
                as_of_times=dates,
                feature_dictionary=feature_dictionary,
                entity_date_table_name=entity_date_table_name,
                matrix_uuid="my_uuid",
            )

            # get the queries and test them
            for result, df in zip(returned_features_dfs, features_dfs):
                test = result == df
                assert test.all().all()


def test_load_labels_data():
    """ Test the load_labels_data function by checking whether the query
    produces the correct labels
    """
    # set up labeling config variables
    dates = [datetime.datetime(2016, 1, 1, 0, 0), datetime.datetime(2016, 2, 1, 0, 0)]

    # make a dataframe of labels to test against
    labels_df = pd.DataFrame(
        labels,
        columns=[
            "entity_id",
            "as_of_date",
            "label_timespan",
            "label_name",
            "label_type",
            "label",
        ],
    )

    labels_df["as_of_date"] = convert_string_column_to_date(labels_df["as_of_date"])
    labels_df.set_index(["entity_id", "as_of_date"])

    # create an engine and generate a table with fake feature data
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_schemas(engine, features_tables, labels, states)
        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                experiment_hash=experiment_hash,
                engine=engine,
            )

            # make the entity-date table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type="binary",
                label_name="booking",
                state="active",
                matrix_type="train",
                matrix_uuid="my_uuid",
                label_timespan="1 month",
            )

            result = builder.load_labels_data(
                label_name=label_name,
                label_type=label_type,
                label_timespan="1 month",
                matrix_uuid="my_uuid",
                entity_date_table_name=entity_date_table_name,
            )
            df = pd.DataFrame.from_dict(
                {
                    "entity_id": [2, 3, 4, 4],
                    "as_of_date": [dates[1], dates[1], dates[0], dates[1]],
                    "booking": [0, 0, 1, 0],
                }
            ).set_index(["entity_id", "as_of_date"])

            test = result == df
            assert test.all().all()


def test_load_labels_data_include_missing_labels_as_false():
    """ Test the load_labels_data function by checking whether the query
    produces the correct labels
    """
    # set up labeling config variables
    dates = [
        datetime.datetime(2016, 1, 1, 0, 0),
        datetime.datetime(2016, 2, 1, 0, 0),
        datetime.datetime(2016, 6, 1, 0, 0),
    ]

    # same as the other load_labels_data test, except we include an extra date, 2016-06-01
    # this date does have entity 0 included via the states table, but no labels

    # make a dataframe of labels to test against
    labels_df = pd.DataFrame(
        labels,
        columns=[
            "entity_id",
            "as_of_date",
            "label_timespan",
            "label_name",
            "label_type",
            "label",
        ],
    )

    labels_df["as_of_date"] = convert_string_column_to_date(labels_df["as_of_date"])
    labels_df.set_index(["entity_id", "as_of_date"])

    # create an engine and generate a table with fake feature data
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_schemas(engine, features_tables, labels, states)
        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                experiment_hash=experiment_hash,
                engine=engine,
                include_missing_labels_in_train_as=False,
            )

            # make the entity-date table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type="binary",
                label_name="booking",
                state="active",
                matrix_type="train",
                matrix_uuid="my_uuid",
                label_timespan="1 month",
            )

            result = builder.load_labels_data(
                label_name=label_name,
                label_type=label_type,
                label_timespan="1 month",
                matrix_uuid="my_uuid",
                entity_date_table_name=entity_date_table_name,
            )
            df = pd.DataFrame.from_dict(
                {
                    "entity_id": [0, 2, 3, 4, 4],
                    "as_of_date": [dates[2], dates[1], dates[1], dates[0], dates[1]],
                    "booking": [0, 0, 0, 1, 0],
                }
            ).set_index(["entity_id", "as_of_date"])
            # the first row would not be here if we had not configured the Builder
            # to include missing labels as false

            test = result == df
            assert test.all().all()


class TestMergeFeatureCSVs(TestCase):
    def test_badinput(self):
        """We assert column names, so replacing 'date' with 'as_of_date'
        should result in an error"""
        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                experiment_hash=experiment_hash,
                engine=None,
            )
            dataframes = [
                pd.DataFrame.from_records(
                    [(1, 3, 3), (4, 5, 6), (7, 8, 9)],
                    columns=("entity_id", "date", "f1"),
                    index=["entity_id", "date"],
                ),
                pd.DataFrame.from_records(
                    [(1, 2, 3), (4, 5, 9), (7, 8, 15)],
                    columns=("entity_id", "date", "f3"),
                    index=["entity_id", "date"],
                ),
                pd.DataFrame.from_records(
                    [(1, 2, 2), (4, 5, 20), (7, 8, 56)],
                    columns=("entity_id", "date", "f3"),
                    index=["entity_id", "date"],
                ),
            ]

            with self.assertRaises(ValueError):
                builder.merge_feature_csvs(dataframes, matrix_uuid="1234")


class TestBuildMatrix(TestCase):
    @property
    def good_metadata(self):
        return {
            "matrix_id": "hi",
            "state": "active",
            "label_name": "booking",
            "end_time": datetime.datetime(2016, 3, 1, 0, 0),
            "feature_start_time": datetime.datetime(2016, 1, 1, 0, 0),
            "label_timespan": "1 month",
            "max_training_history": "1 month",
            "test_duration": "1 month",
            "indices": ["entity_id", "as_of_date"],
        }

    @property
    def good_feature_dictionary(self):
        return FeatureGroup(
            name="mygroup",
            features_by_table={"features0": ["f1", "f2"], "features1": ["f3", "f4"]},
        )

    @property
    def good_dates(self):
        return [
            datetime.datetime(2016, 1, 1, 0, 0),
            datetime.datetime(2016, 2, 1, 0, 0),
            datetime.datetime(2016, 3, 1, 0, 0),
        ]

    def test_train_matrix(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                )
                uuid = filename_friendly_hash(self.good_metadata)
                builder.build_matrix(
                    as_of_times=self.good_dates,
                    label_name="booking",
                    label_type="binary",
                    feature_dictionary=self.good_feature_dictionary,
                    matrix_metadata=self.good_metadata,
                    matrix_uuid=uuid,
                    matrix_type="train",
                )
                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5
                assert builder.sessionmaker().query(Matrix).get(uuid).feature_dictionary ==self.good_feature_dictionary

    def test_test_matrix(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                )

                uuid = filename_friendly_hash(self.good_metadata)
                builder.build_matrix(
                    as_of_times=self.good_dates,
                    label_name="booking",
                    label_type="binary",
                    feature_dictionary=self.good_feature_dictionary,
                    matrix_metadata=self.good_metadata,
                    matrix_uuid=uuid,
                    matrix_type="test",
                )

                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5

    def test_nullcheck(self):
        f0_dict = {(r[0], r[1]): r for r in features0_pre}
        f1_dict = {(r[0], r[1]): r for r in features1_pre}

        features0 = sorted(f0_dict.values(), key=lambda x: (x[1], x[0]))
        features1 = sorted(f1_dict.values(), key=lambda x: (x[1], x[0]))

        features_tables = [features0, features1]

        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )

            dates = [
                datetime.datetime(2016, 1, 1, 0, 0),
                datetime.datetime(2016, 2, 1, 0, 0),
                datetime.datetime(2016, 3, 1, 0, 0),
            ]

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                )

                feature_dictionary = {
                    "features0": ["f1", "f2"],
                    "features1": ["f3", "f4"],
                }
                matrix_metadata = {
                    "matrix_id": "hi",
                    "state": "active",
                    "label_name": "booking",
                    "end_time": datetime.datetime(2016, 3, 1, 0, 0),
                    "feature_start_time": datetime.datetime(2016, 1, 1, 0, 0),
                    "label_timespan": "1 month",
                    "test_duration": "1 month",
                    "indices": ["entity_id", "as_of_date"],
                }
                uuid = filename_friendly_hash(matrix_metadata)
                with self.assertRaises(ValueError):
                    builder.build_matrix(
                        as_of_times=dates,
                        label_name="booking",
                        label_type="binary",
                        feature_dictionary=feature_dictionary,
                        matrix_metadata=matrix_metadata,
                        matrix_uuid=uuid,
                        matrix_type="test",
                    )

    def test_replace_false_rerun(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )

            dates = [
                datetime.datetime(2016, 1, 1, 0, 0),
                datetime.datetime(2016, 2, 1, 0, 0),
                datetime.datetime(2016, 3, 1, 0, 0),
            ]

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                    replace=False,
                )

                feature_dictionary = {
                    "features0": ["f1", "f2"],
                    "features1": ["f3", "f4"],
                }
                matrix_metadata = {
                    "matrix_id": "hi",
                    "state": "active",
                    "label_name": "booking",
                    "end_time": datetime.datetime(2016, 3, 1, 0, 0),
                    "feature_start_time": datetime.datetime(2016, 1, 1, 0, 0),
                    "label_timespan": "1 month",
                    "test_duration": "1 month",
                    "indices": ["entity_id", "as_of_date"],
                }
                uuid = filename_friendly_hash(matrix_metadata)
                builder.build_matrix(
                    as_of_times=dates,
                    label_name="booking",
                    label_type="binary",
                    feature_dictionary=feature_dictionary,
                    matrix_metadata=matrix_metadata,
                    matrix_uuid=uuid,
                    matrix_type="test",
                )

                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5
                # rerun
                builder.make_entity_date_table = Mock()
                builder.build_matrix(
                    as_of_times=dates,
                    label_name="booking",
                    label_type="binary",
                    feature_dictionary=feature_dictionary,
                    matrix_metadata=matrix_metadata,
                    matrix_uuid=uuid,
                    matrix_type="test",
                )
                assert not builder.make_entity_date_table.called

    def test_replace_true_rerun(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )
            matrix_metadata = matrix_metadata_creator(
                state="active", test_duration="1month", label_name="booking"
            )

            dates = [
                datetime.datetime(2016, 1, 1, 0, 0),
                datetime.datetime(2016, 2, 1, 0, 0),
                datetime.datetime(2016, 3, 1, 0, 0),
            ]

            feature_dictionary = {"features0": ["f1", "f2"], "features1": ["f3", "f4"]}
            uuid = filename_friendly_hash(matrix_metadata)
            build_args = dict(
                as_of_times=dates,
                label_name="booking",
                label_type="binary",
                feature_dictionary=feature_dictionary,
                matrix_metadata=matrix_metadata,
                matrix_uuid=uuid,
                matrix_type="test",
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                    replace=True,
                )

                builder.build_matrix(**build_args)

                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5
                assert builder.sessionmaker().query(Matrix).get(uuid)
                # rerun
                builder.build_matrix(**build_args)
                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5
                assert builder.sessionmaker().query(Matrix).get(uuid)



"""empty message

Revision ID: 7d57d1cf3429
Revises: 72ac5cbdca05
Create Date: 2017-11-06 11:34:23.046005

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "7d57d1cf3429"
down_revision = "72ac5cbdca05"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        "evaluations", "example_frequency", new_column_name="as_of_date_frequency"
    )
    op.alter_column(
        "models", "train_label_window", new_column_name="training_label_timespan"
    )
    op.alter_column(
        "predictions", "test_label_window", new_column_name="test_label_timespan"
    )
    op.alter_column(
        "list_predictions", "test_label_window", new_column_name="test_label_timespan"
    )


def downgrade():
    op.alter_column(
        "evaluations", "as_of_date_frequency", new_column_name="example_frequency"
    )
    op.alter_column(
        "models", "training_label_timespan", new_column_name="train_label_window"
    )
    op.alter_column(
        "predictions", "test_label_timespan", new_column_name="test_label_window"
    )
    op.alter_column(
        "list_predictions", "test_label_timespan", new_column_name="test_label_window"
    )



# Avoid circular import (required by base)
CONFIG_VERSION = "v7"  # noqa: E402

from .base import ExperimentBase
from .multicore import MultiCoreExperiment
from .singlethreaded import SingleThreadedExperiment

__all__ = ("ExperimentBase", "MultiCoreExperiment", "SingleThreadedExperiment")



# -*- coding: utf-8 -*-

__author__ = """Center for Data Science and Public Policy"""
__email__ = "datascifellows@gmail.com"
__version__ = '4.4.0' # do not change to double-quotes, it will screw up bumpversion

import logging
import logging.config
import yaml
import pathlib


logging_config = pathlib.Path(__file__).parent / 'config' / 'logging.yaml'

with open(logging_config, 'r') as f:
    config = yaml.safe_load(f.read())
    logging.config.dictConfig(config)


from .util.db import create_engine



__all__ = ('create_engine',)
